#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import dataclasses
from typing import cast

from pydantic import ValidationError
from quart import render_template

from otterdog.webapp.tasks import Task
from otterdog.webapp.webhook.github_models import PullRequest, Repository


@dataclasses.dataclass(repr=False)
class RetrieveTeamMembershipTask(Task[None]):
    installation_id: int
    org_id: str
    repository: Repository
    pull_request_or_number: PullRequest | int

    async def _pre_execute(self) -> None:
        rest_api = await self.get_rest_api(self.installation_id)

        if isinstance(self.pull_request_or_number, int):
            response = await rest_api.pull_request.get_pull_request(
                self.org_id, self.repository.name, str(self.pull_request_or_number)
            )
            try:
                self.pull_request = PullRequest.model_validate(response)
            except ValidationError as ex:
                self.logger.exception("failed to load pull request event data", exc_info=ex)
                return
        else:
            self.pull_request = cast(PullRequest, self.pull_request_or_number)

        self.logger.info(
            "retrieving team membership of author '%s' for pull request #%d of repo '%s'",
            self.pull_request.user.login,
            self.pull_request.number,
            self.repository.full_name,
        )

    async def _execute(self) -> None:
        rest_api = await self.get_rest_api(self.installation_id)

        user = self.pull_request.user.login
        team_slugs = await rest_api.team.get_team_slugs(self.org_id)
        team_membership = []

        for team_slug in team_slugs:
            if await rest_api.team.is_user_member_of_team(self.org_id, team_slug, user):
                team_membership.append(team_slug)

        teams = [(team, f"https://github.com/orgs/{self.org_id}/teams/{team}") for team in team_membership]
        comment = await render_template("team_membership_comment.txt", user=user, teams=teams)
        await rest_api.issue.create_comment(self.org_id, self.repository.name, str(self.pull_request.number), comment)

    def __repr__(self) -> str:
        pull_request_number = (
            self.pull_request_or_number
            if isinstance(self.pull_request_or_number, int)
            else self.pull_request_or_number.number
        )
        return f"RetrieveTeamMembershipTask(repo={self.repository.full_name}, pull_request={pull_request_number})"



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import os
from abc import abstractmethod
from typing import Any, Optional, Protocol

from otterdog.config import OrganizationConfig, OtterdogConfig
from otterdog.jsonnet import JsonnetConfig
from otterdog.models import LivePatch, LivePatchContext, LivePatchType, ModelObject
from otterdog.models.github_organization import GitHubOrganization
from otterdog.providers.github import GitHubProvider
from otterdog.utils import Change, IndentingPrinter, style

from . import Operation
from .validate import ValidateOperation


class DiffStatus:
    def __init__(self):
        self.additions = 0
        self.differences = 0
        self.deletions = 0

    def total_changes(self, include_deletions: bool) -> int:
        if include_deletions:
            return self.additions + self.differences + self.deletions
        else:
            return self.additions + self.differences


class CallbackFn(Protocol):
    def __call__(self, org_id: str, diff_status: DiffStatus, patches: list[LivePatch]) -> None:
        ...


class DiffOperation(Operation):
    def __init__(self, no_web_ui: bool, update_webhooks: bool, update_secrets: bool, update_filter: str):
        super().__init__()

        self.no_web_ui = no_web_ui
        self.update_webhooks = update_webhooks
        self.update_secrets = update_secrets
        self.update_filter = update_filter
        self._gh_client: Optional[GitHubProvider] = None
        self._validator = ValidateOperation()
        self._template_dir: Optional[str] = None
        self._org_config: Optional[OrganizationConfig] = None
        self._callback: Optional[CallbackFn] = None

    @property
    def template_dir(self) -> str:
        assert self._template_dir is not None
        return self._template_dir

    @property
    def org_config(self) -> OrganizationConfig:
        assert self._org_config is not None
        return self._org_config

    def set_callback(self, fn: CallbackFn) -> None:
        self._callback = fn

    def init(self, config: OtterdogConfig, printer: IndentingPrinter) -> None:
        super().init(config, printer)
        self._validator.init(config, printer)

    async def execute(self, org_config: OrganizationConfig) -> int:
        self._org_config = org_config

        self.printer.println(f"\nOrganization {style(org_config.name, bright=True)}[id={org_config.github_id}]")

        try:
            self._gh_client = self.setup_github_client(org_config)
        except RuntimeError as e:
            self.printer.print_error(f"invalid credentials\n{str(e)}")
            return 1

        self.printer.level_up()

        try:
            return await self._generate_diff(org_config)
        finally:
            self.printer.level_down()
            self._gh_client.close()

    def setup_github_client(self, org_config: OrganizationConfig) -> GitHubProvider:
        return GitHubProvider(self.config.get_credentials(org_config, only_token=self.no_web_ui))

    @property
    def gh_client(self) -> GitHubProvider:
        assert self._gh_client is not None
        return self._gh_client

    def verbose_output(self):
        return True

    def include_resources_with_secrets(self) -> bool:
        return True

    def resolve_secrets(self) -> bool:
        return True

    async def _generate_diff(self, org_config: OrganizationConfig) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config
        jsonnet_config.init_template()
        self._template_dir = jsonnet_config.template_dir

        org_file_name = jsonnet_config.org_config_file

        if not os.path.exists(org_file_name):
            self.printer.print_error(
                f"configuration file '{org_file_name}' does not yet exist, run fetch-config or import first."
            )
            return 1

        try:
            expected_org = self.load_expected_org(github_id, org_file_name)
        except RuntimeError as e:
            self.printer.print_error(f"failed to load configuration\n{str(e)}")
            return 1

        # We validate the configuration first and only calculate a plan if
        # there are no validation errors.
        (
            validation_infos,
            validation_warnings,
            validation_errors,
        ) = self._validator.validate(expected_org, jsonnet_config.template_dir)
        if validation_errors > 0:
            self.printer.println("Planning aborted due to validation errors.")
            return validation_errors

        if validation_infos > 0 and not self.printer.is_info_enabled():
            self.printer.println(
                f"there have been {validation_infos} validation infos, "
                f"enable verbose output with '-v' to to display them."
            )

        try:
            current_org = await self.load_current_org(github_id, jsonnet_config)
        except RuntimeError as e:
            self.printer.print_error(f"failed to load current configuration\n{str(e)}")
            return 1

        diff_status = DiffStatus()
        live_patches = []

        def handle(patch: LivePatch) -> None:
            if not self.include_resources_with_secrets() and patch.requires_secrets():
                return

            live_patches.append(patch)

            match patch.patch_type:
                case LivePatchType.ADD:
                    assert patch.expected_object is not None
                    self.handle_add_object(github_id, patch.expected_object, patch.parent_object)
                    diff_status.additions += 1

                case LivePatchType.REMOVE:
                    assert patch.current_object is not None
                    self.handle_delete_object(github_id, patch.current_object, patch.parent_object)
                    diff_status.deletions += 1

                case LivePatchType.CHANGE:
                    assert patch.changes is not None
                    assert patch.current_object is not None
                    assert patch.expected_object is not None
                    diff_status.differences += self.handle_modified_object(
                        github_id,
                        patch.changes,
                        False,
                        patch.current_object,
                        patch.expected_object,
                        patch.parent_object,
                    )

        context = LivePatchContext(
            github_id, self.update_webhooks, self.update_secrets, self.update_filter, expected_org.settings
        )
        expected_org.generate_live_patch(current_org, context, handle)

        # add a warning that otterdog potentially must be run a second time
        # to fully apply all settings.
        if "web_commit_signoff_required" in context.modified_org_settings:
            change = context.modified_org_settings["web_commit_signoff_required"]
            if change.to_value is False:
                if self.verbose_output():
                    self.printer.print_warn(
                        "Setting 'web_commit_signoff_required' setting has been disabled on "
                        "organization level. \nThe effective setting on repo level can only be "
                        "determined once this change has been applied.\n"
                        "You need to run otterdog another time to fully ensure "
                        "that the correct configuration is applied."
                    )

        # resolve secrets for collected patches
        if self.resolve_secrets():
            for live_patch in live_patches:
                if live_patch.expected_object is not None:
                    live_patch.expected_object.resolve_secrets(self.config.get_secret)

        await self.handle_finish(github_id, diff_status, live_patches)

        if self._callback is not None:
            self._callback(github_id, diff_status, live_patches)

        return 0

    def load_expected_org(self, github_id: str, org_file_name: str) -> GitHubOrganization:
        return GitHubOrganization.load_from_file(github_id, org_file_name, self.config)

    async def load_current_org(self, github_id: str, jsonnet_config: JsonnetConfig) -> GitHubOrganization:
        return await GitHubOrganization.load_from_provider(
            github_id, jsonnet_config, self.gh_client, self.no_web_ui, self.printer
        )

    @abstractmethod
    def handle_add_object(
        self,
        org_id: str,
        model_object: ModelObject,
        parent_object: Optional[ModelObject] = None,
    ) -> None:
        ...

    @abstractmethod
    def handle_delete_object(
        self,
        org_id: str,
        model_object: ModelObject,
        parent_object: Optional[ModelObject] = None,
    ) -> None:
        ...

    @abstractmethod
    def handle_modified_object(
        self,
        org_id: str,
        modified_object: dict[str, Change[Any]],
        forced_update: bool,
        current_object: ModelObject,
        expected_object: ModelObject,
        parent_object: Optional[ModelObject] = None,
    ) -> int:
        ...

    @abstractmethod
    async def handle_finish(self, org_id: str, diff_status: DiffStatus, patches: list[LivePatch]) -> None:
        ...



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from typing import Any, MutableMapping, Optional

from jwt import JWT, jwk_from_pem
from jwt.utils import get_int_from_datetime

from . import AuthImpl, AuthStrategy


@dataclass(frozen=True)
class AppAuthStrategy(AuthStrategy):
    """
    A strategy to authenticate as a GitHub App.
    """

    app_id: str
    private_key: str

    def get_auth(self) -> AuthImpl:
        return _AppAuth(self.app_id, self.private_key)


@dataclass
class _AppAuth(AuthImpl):
    app_id: str
    private_key: str

    _jwt: Optional[str] = None
    _expire: Optional[datetime] = None

    def _create_jwt(self) -> tuple[str, datetime, datetime]:
        """
        Create a JWT authenticating as a GitHub App. See
        https://docs.github.com/en/developers/apps/building-github-apps/authenticating-with-github-apps#authenticating-as-a-github-app
        """

        # Open PEM
        with open(self.private_key, "rb") as pem_file:
            signing_key = jwk_from_pem(pem_file.read())

        # use a start time slightly in the past
        start_time = datetime.now(timezone.utc) - timedelta(seconds=10)
        # JWT expiration time (10 minutes maximum)
        expire_time = start_time + timedelta(minutes=10)

        payload = {
            "iat": get_int_from_datetime(start_time),
            "exp": get_int_from_datetime(expire_time),
            "iss": self.app_id,
        }

        return JWT().encode(payload, signing_key, alg="RS256"), start_time, expire_time

    def get_jwt(self) -> str:
        now = datetime.now(timezone.utc)

        if self._jwt is None or self._expire is None or now > self._expire:
            self._jwt, _, self._expire = self._create_jwt()

        return self._jwt

    def __call__(self, r):
        self.update_headers_with_authorization(r.headers)
        return r

    def update_headers_with_authorization(self, headers: MutableMapping[str, Any]) -> None:
        headers["Authorization"] = f"Bearer {self.get_jwt()}"



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import json
import os
import pathlib
import re
import tempfile
import zipfile
from typing import IO, Any, Optional

import chevron
import jq  # type: ignore

from otterdog.providers.github.exception import GitHubException
from otterdog.providers.github.rest import RestApi, RestClient, encrypt_value
from otterdog.utils import (
    associate_by_key,
    is_set_and_present,
    print_debug,
    print_trace,
)


class RepoClient(RestClient):
    def __init__(self, rest_api: RestApi):
        super().__init__(rest_api)

    async def get_repo_data(self, org_id: str, repo_name: str) -> dict[str, Any]:
        print_debug(f"retrieving org repo data for '{org_id}/{repo_name}'")

        try:
            repo_data = await self.requester.async_request_json("GET", f"/repos/{org_id}/{repo_name}")
            await self._fill_github_pages_config(org_id, repo_name, repo_data)
            await self._fill_vulnerability_report(org_id, repo_name, repo_data)
            await self._fill_topics(org_id, repo_name, repo_data)
            return repo_data
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving data for repo '{repo_name}':\n{ex}").with_traceback(tb)

    async def get_repo_by_id(self, repo_id: int) -> dict[str, Any]:
        print_debug(f"retrieving repo by id for '{repo_id}'")

        try:
            return await self.requester.async_request_json("GET", f"/repositories/{repo_id}")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving data for repo '{repo_id}':\n{ex}").with_traceback(tb)

    async def update_repo(self, org_id: str, repo_name: str, data: dict[str, Any]) -> None:
        print_debug(f"updating repo settings for repo '{org_id}/{repo_name}'")

        changes = len(data)

        if "dependabot_alerts_enabled" in data:
            vulnerability_reports = bool(data.pop("dependabot_alerts_enabled"))
        else:
            vulnerability_reports = None

        if "topics" in data:
            topics = list(data.pop("topics"))
        else:
            topics = None

        if "gh_pages" in data:
            gh_pages = data.pop("gh_pages")
        else:
            gh_pages = None

        if "default_branch" in data:
            default_branch = data.pop("default_branch")
        else:
            default_branch = None

        if changes > 0:
            try:
                if len(data) > 0:
                    await self.requester.async_request_json("PATCH", f"/repos/{org_id}/{repo_name}", data)

                if vulnerability_reports is not None:
                    await self._update_vulnerability_report(org_id, repo_name, vulnerability_reports)
                if topics is not None:
                    await self._update_topics(org_id, repo_name, topics)
                if gh_pages is not None:
                    await self._update_github_pages_config(org_id, repo_name, gh_pages)
                if default_branch is not None:
                    await self._update_default_branch(org_id, repo_name, default_branch)

                print_debug(f"updated {changes} repo setting(s) for repo '{repo_name}'")
            except GitHubException as ex:
                tb = ex.__traceback__
                raise RuntimeError(f"failed to update settings for repo '{repo_name}':\n{ex}").with_traceback(tb)

    async def add_repo(
        self,
        org_id: str,
        data: dict[str, Any],
        template_repository: Optional[str],
        post_process_template_content: list[str],
        forked_repository: Optional[str],
        fork_default_branch_only: bool,
        auto_init_repo: bool,
    ) -> None:
        repo_name = data["name"]

        if is_set_and_present(forked_repository):
            print_debug(f"forking repo '{forked_repository}' to '{org_id}/{repo_name}'")
            upstream_owner, upstream_repo = re.split("/", forked_repository, 1)

            try:
                fork_data = {
                    "organization": org_id,
                    "name": repo_name,
                    "default_branch_only": fork_default_branch_only,
                }

                await self.requester.async_request_json(
                    "POST",
                    f"/repos/{upstream_owner}/{upstream_repo}/forks",
                    fork_data,
                )

                # get all the data for the created repo to avoid setting values that can not be changed due
                # to defaults from the organization (like web_commit_signoff_required)
                current_data = await self.get_repo_data(org_id, repo_name)
                self._remove_already_active_settings(data, current_data)
                await self.update_repo(org_id, repo_name, data)

                print_debug(f"created repo with name '{repo_name}' from template '{template_repository}'")
                return
            except GitHubException as ex:
                tb = ex.__traceback__
                raise RuntimeError(
                    f"failed to fork repo '{repo_name}' from repo '{forked_repository}':\n{ex}"
                ).with_traceback(tb)

        if is_set_and_present(template_repository):
            print_debug(f"creating repo '{org_id}/{repo_name}' with template '{template_repository}'")
            template_owner, template_repo = re.split("/", template_repository, 1)

            try:
                template_data = {
                    "owner": org_id,
                    "name": repo_name,
                    "include_all_branches": False,
                    "private": data.get("private", False),
                }

                await self.requester.async_request_json(
                    "POST",
                    f"/repos/{template_owner}/{template_repo}/generate",
                    template_data,
                )

                print_debug(f"created repo with name '{repo_name}' from template '{template_repository}'")

                # get all the data for the created repo to avoid setting values that can not be changed due
                # to defaults from the organization (like web_commit_signoff_required)
                current_data = await self.get_repo_data(org_id, repo_name)
                self._remove_already_active_settings(data, current_data)
                await self.update_repo(org_id, repo_name, data)

                # wait till the repo is initialized, this might take a while.
                if len(post_process_template_content) > 0:
                    initialized = False
                    for i in range(1, 11):
                        try:
                            await self.get_readme(org_id, repo_name)
                            initialized = True
                            break
                        except RuntimeError:
                            print_trace(f"waiting for repo '{org_id}/{repo_name}' to be initialized, " f"try {i} of 10")
                            import time

                            time.sleep(1)

                    if initialized is False:
                        raise RuntimeError(
                            f"failed to create repo from template '{template_repository}': "
                            f"repo not initialized after 5s"
                        )

                # if there is template content which shall be post-processed,
                # use chevron to expand some variables that might be used there.
                for content_path in post_process_template_content:
                    content = await self.rest_api.content.get_content(org_id, repo_name, content_path, None)
                    updated_content = self._render_template_content(org_id, repo_name, content)
                    if content != updated_content:
                        await self.rest_api.content.update_content(org_id, repo_name, content_path, updated_content)

                return
            except GitHubException as ex:
                tb = ex.__traceback__
                raise RuntimeError(
                    f"failed to create repo from template '{template_repository}':\n{ex}"
                ).with_traceback(tb)

        print_debug(f"creating repo '{org_id}/{repo_name}'")

        # some settings do not seem to be set correctly during creation
        # collect them and update the repo after creation.
        update_keys = [
            "dependabot_alerts_enabled",
            "web_commit_signoff_required",
            "security_and_analysis",
            "topics",
            "gh_pages",
        ]

        if auto_init_repo is True:
            update_keys.append("default_branch")

        update_data = {}

        for update_key in update_keys:
            if update_key in data:
                update_data[update_key] = data.pop(update_key)

        # whether the repo should be initialized with an empty README
        data["auto_init"] = auto_init_repo

        try:
            result = await self.requester.async_request_json("POST", f"/orgs/{org_id}/repos", data)
            print_debug(f"created repo with name '{repo_name}'")
            self._remove_already_active_settings(update_data, result)
            await self.update_repo(org_id, repo_name, update_data)
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed to add repo with name '{org_id}/{repo_name}':\n{ex}").with_traceback(tb)

    async def get_webhooks(self, org_id: str, repo_name: str) -> list[dict[str, Any]]:
        print_debug(f"retrieving webhooks for repo '{org_id}/{repo_name}'")

        try:
            return await self.requester.async_request_json("GET", f"/repos/{org_id}/{repo_name}/hooks")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving webhooks for repo '{org_id}/{repo_name}':\n{ex}").with_traceback(tb)

    async def update_webhook(self, org_id: str, repo_name: str, webhook_id: int, webhook: dict[str, Any]) -> None:
        print_debug(f"updating repo webhook '{webhook_id}' for repo '{org_id}/{repo_name}'")

        try:
            await self.requester.async_request_json("PATCH", f"/repos/{org_id}/{repo_name}/hooks/{webhook_id}", webhook)
            print_debug(f"updated repo webhook '{webhook_id}'")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed to update repo webhook {webhook_id}:\n{ex}").with_traceback(tb)

    async def add_webhook(self, org_id: str, repo_name: str, data: dict[str, Any]) -> None:
        url = data["config"]["url"]
        print_debug(f"adding repo webhook with url '{url}' for repo '{org_id}/{repo_name}'")

        # mandatory field "name" = "web"
        data["name"] = "web"

        try:
            await self.requester.async_request_json("POST", f"/repos/{org_id}/{repo_name}/hooks", data)
            print_debug(f"added repo webhook with url '{url}'")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed to add repo webhook with url '{url}':\n{ex}").with_traceback(tb)

    async def delete_webhook(self, org_id: str, repo_name: str, webhook_id: int, url: str) -> None:
        print_debug(f"deleting repo webhook with url '{url}' for repo '{org_id}/{repo_name}'")

        status, _ = await self.requester.async_request_raw("DELETE", f"/repos/{org_id}/{repo_name}/hooks/{webhook_id}")

        if status != 204:
            raise RuntimeError(f"failed to delete repo webhook with url '{url}'")

        print_debug(f"removed repo webhook with url '{url}'")

    async def get_rulesets(self, org_id: str, repo_name: str) -> list[dict[str, Any]]:
        print_debug(f"retrieving rulesets for repo '{org_id}/{repo_name}'")

        try:
            result = []
            params = {"includes_parents": str(False)}
            response = await self.requester.async_request_paged_json(
                "GET", f"/repos/{org_id}/{repo_name}/rulesets", params=params
            )
            for ruleset in response:
                result.append(await self.get_ruleset(org_id, repo_name, str(ruleset["id"])))
            return result
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving rulesets for repo '{org_id}/{repo_name}':\n{ex}").with_traceback(tb)

    async def get_ruleset(self, org_id: str, repo_name: str, ruleset_id: str) -> dict[str, Any]:
        print_debug(f"retrieving ruleset '{ruleset_id}' for repo '{org_id}/{repo_name}'")

        try:
            params = {"includes_parents": str(False)}
            return await self.requester.async_request_json(
                "GET", f"/repos/{org_id}/{repo_name}/rulesets/{ruleset_id}", params=params
            )
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving ruleset for repo '{org_id}/{repo_name}':\n{ex}").with_traceback(tb)

    async def update_ruleset(self, org_id: str, repo_name: str, ruleset_id: int, ruleset: dict[str, Any]) -> None:
        print_debug(f"updating repo ruleset '{ruleset_id}' for repo '{org_id}/{repo_name}'")

        try:
            await self.requester.async_request_json(
                "PUT", f"/repos/{org_id}/{repo_name}/rulesets/{ruleset_id}", ruleset
            )
            print_debug(f"updated repo ruleset '{ruleset_id}'")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed to update repo ruleset {ruleset_id}:\n{ex}").with_traceback(tb)

    async def add_ruleset(self, org_id: str, repo_name: str, data: dict[str, Any]) -> None:
        name = data["name"]
        print_debug(f"adding repo ruleset with name '{name}' for repo '{org_id}/{repo_name}'")

        # TODO: currently we only support rulesets targetting branches
        data["target"] = "branch"

        try:
            await self.requester.async_request_json("POST", f"/repos/{org_id}/{repo_name}/rulesets", data)
            print_debug(f"added repo ruleset with name '{name}'")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed to add repo ruleset with name '{name}':\n{ex}").with_traceback(tb)

    async def delete_ruleset(self, org_id: str, repo_name: str, ruleset_id: int, name: str) -> None:
        print_debug(f"deleting repo ruleset with name '{name}' for repo '{org_id}/{repo_name}'")

        status, _ = await self.requester.async_request_raw(
            "DELETE", f"/repos/{org_id}/{repo_name}/rulesets/{ruleset_id}"
        )

        if status != 204:
            raise RuntimeError(f"failed to delete repo ruleset with name '{name}'")

        print_debug(f"removed repo ruleset with name '{name}'")

    @staticmethod
    def _render_template_content(org_id: str, repo_name: str, content: str) -> str:
        variables = {"org": org_id, "repo": repo_name}
        return chevron.render(content, variables)

    async def get_readme(self, org_id: str, repo_name: str) -> dict[str, Any]:
        print_debug(f"getting readme for repo '{org_id}/{repo_name}'")

        try:
            return await self.requester.async_request_json("GET", f"/repos/{org_id}/{repo_name}/readme")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed to get readme for repo '{org_id}/{repo_name}':\n{ex}").with_traceback(tb)

    async def delete_repo(self, org_id: str, repo_name: str) -> None:
        print_debug(f"deleting repo '{org_id}/{repo_name}'")

        status, body = await self.requester.async_request_raw("DELETE", f"/repos/{org_id}/{repo_name}")

        if status != 204:
            raise RuntimeError(f"failed to delete repo '{org_id}/{repo_name}': {body}")

        print_debug(f"removed repo '{org_id}/{repo_name}'")

    @staticmethod
    def _remove_already_active_settings(update_data: dict[str, Any], current_data: dict[str, Any]) -> None:
        keys = list(update_data.keys())
        for key in keys:
            if key in current_data:
                update_value_expected = update_data[key]
                update_value_current = current_data[key]

                if update_value_current == update_value_expected:
                    print_debug(f"omitting setting '{key}' as it is already set")
                    update_data.pop(key)

    async def _fill_github_pages_config(self, org_id: str, repo_name: str, repo_data: dict[str, Any]) -> None:
        print_debug(f"retrieving github pages config for '{org_id}/{repo_name}'")

        status, body = await self.requester.async_request_raw("GET", f"/repos/{org_id}/{repo_name}/pages")
        if status == 200:
            repo_data["gh_pages"] = json.loads(body)

    async def _update_github_pages_config(self, org_id: str, repo_name: str, gh_pages: dict[str, Any]) -> None:
        print_debug(f"updating github pages config for '{org_id}/{repo_name}'")

        # special handling for repos hosting the organization site
        if repo_name.lower() == f"{org_id}.github.io".lower():
            current_repo_data: dict[str, Any] = {}
            for i in range(1, 4):
                await self._fill_github_pages_config(org_id, repo_name, current_repo_data)
                if "gh_pages" in current_repo_data:
                    break

                print_trace(f"waiting for repo '{org_id}/{repo_name}' to be initialized, " f"try {i} of 3")
                import time

                time.sleep(1)

            current_gh_pages: Any = current_repo_data.get("gh_pages", None)
            if current_gh_pages is not None:
                has_changes = False
                for k, v in gh_pages.items():
                    if current_gh_pages.get(k, None) != v:
                        has_changes = True
                        break

                # if there are no changes to the current config, we do not need to do anything
                if has_changes is False:
                    print_trace(f"github pages config for '{org_id}/{repo_name}' is already up-to-date")
                    return

        build_type = gh_pages.get("build_type", None)
        if build_type == "disabled":
            self.requester.request_raw("DELETE", f"/repos/{org_id}/{repo_name}/pages")
        else:
            gh_pages_data: list[tuple[str, str, int]] = []
            # first check if the pages config already exists:
            status, _ = await self.requester.async_request_raw("GET", f"/repos/{org_id}/{repo_name}/pages")
            if status != 200:
                # check if the branch already exists
                source: Any = gh_pages.get("source", None)
                if source is not None:
                    branch = source.get("branch", None)
                    if branch is not None:
                        existing_branches = await self.get_branches(org_id, repo_name)

                        if len(existing_branches) == 0:
                            print_debug(f"repo '{repo_name}' not yet initialized, skipping GH pages config")
                            return

                        existing_branch_names = list(map(lambda x: x["name"], existing_branches))
                        if branch not in existing_branch_names:
                            gh_pages_data.append((json.dumps(gh_pages), "PUT", 204))
                            gh_pages["source"]["branch"] = existing_branch_names[0]
                            gh_pages_data.insert(0, (json.dumps(gh_pages), "POST", 201))

                if len(gh_pages_data) == 0:
                    gh_pages_data.append((json.dumps(gh_pages), "POST", 201))
            else:
                gh_pages_data.append((json.dumps(gh_pages), "PUT", 204))

            for data, method, status_code in gh_pages_data:
                status, body = await self.requester.async_request_raw(
                    method, f"/repos/{org_id}/{repo_name}/pages", data=data
                )

                if status != status_code:
                    raise RuntimeError(f"failed to update github pages config for repo '{repo_name}': {body}")

                print_debug(f"updated github pages config for repo '{repo_name}'")

    async def _update_default_branch(self, org_id: str, repo_name: str, new_default_branch: str) -> None:
        print_debug(f"updating default branch for '{org_id}/{repo_name}'")
        existing_branches = await self.get_branches(org_id, repo_name)
        existing_branch_names = list(map(lambda x: x["name"], existing_branches))

        if len(existing_branches) == 0:
            print_debug(f"skip updating of default branch for empty repo '{org_id}/{repo_name}'")
            return

        try:
            if new_default_branch in existing_branch_names:
                data = {"default_branch": new_default_branch}
                await self.requester.async_request_json("PATCH", f"/repos/{org_id}/{repo_name}", data)
                print_debug(f"updated default branch for '{org_id}/{repo_name}'")
            else:
                repo = await self.get_repo_data(org_id, repo_name)
                default_branch = repo["default_branch"]
                data = {"new_name": new_default_branch}
                await self.requester.async_request_json(
                    "POST", f"/repos/{org_id}/{repo_name}/branches/{default_branch}/rename", data
                )
                print_debug(f"renamed default branch for '{org_id}/{repo_name}'")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(
                f"failed to update default branch for repo '{org_id}/{repo_name}':\n{ex}"
            ).with_traceback(tb)

    async def _fill_vulnerability_report(self, org_id: str, repo_name: str, repo_data: dict[str, Any]) -> None:
        print_debug(f"retrieving repo vulnerability report status for '{org_id}/{repo_name}'")

        status, _ = await self.requester.async_request_raw("GET", f"/repos/{org_id}/{repo_name}/vulnerability-alerts")
        if status == 204:
            repo_data["dependabot_alerts_enabled"] = True
        else:
            repo_data["dependabot_alerts_enabled"] = False

    async def _update_vulnerability_report(self, org_id: str, repo_name: str, vulnerability_reports: bool) -> None:
        print_debug(f"updating repo vulnerability report status for '{org_id}/{repo_name}'")

        if vulnerability_reports is True:
            method = "PUT"
        else:
            method = "DELETE"

        status, body = await self.requester.async_request_raw(
            method, f"/repos/{org_id}/{repo_name}/vulnerability-alerts"
        )

        if status != 204:
            raise RuntimeError(f"failed to update vulnerability_reports for repo '{repo_name}': {body}")

        print_debug(f"updated vulnerability_reports for repo '{repo_name}'")

    async def _fill_topics(self, org_id: str, repo_name: str, repo_data: dict[str, Any]) -> None:
        print_debug(f"retrieving repo topics for '{org_id}/{repo_name}'")

        try:
            response = await self.requester.async_request_json("GET", f"/repos/{org_id}/{repo_name}/topics")
            repo_data["topics"] = response.get("names", [])
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving topics for repo '{org_id}/{repo_name}':\n{ex}").with_traceback(tb)

    async def _update_topics(self, org_id: str, repo_name: str, topics: list[str]) -> None:
        print_debug(f"updating repo topics for '{org_id}/{repo_name}'")
        data = {"names": topics}
        await self.requester.async_request_json("PUT", f"/repos/{org_id}/{repo_name}/topics", data=data)
        print_debug(f"updated topics for repo '{repo_name}'")

    async def get_branches(self, org_id: str, repo_name) -> list[dict[str, Any]]:
        print_debug(f"retrieving branches for repo '{org_id}/{repo_name}'")

        try:
            return await self.requester.async_request_json("GET", f"/repos/{org_id}/{repo_name}/branches")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed getting branches for repo '{org_id}/{repo_name}':\n{ex}").with_traceback(tb)

    async def get_environments(self, org_id: str, repo_name: str) -> list[dict[str, Any]]:
        print_debug(f"retrieving environments for repo '{org_id}/{repo_name}'")

        try:
            response = await self.requester.async_request_json("GET", f"/repos/{org_id}/{repo_name}/environments")

            environments = response["environments"]
            for env in environments:
                env_name = env["name"]
                has_branch_policies = (
                    jq.compile(".deployment_branch_policy.custom_branch_policies // false").input(env).first()
                )

                if has_branch_policies:
                    env["branch_policies"] = await self._get_deployment_branch_policies(org_id, repo_name, env_name)
            return environments
        except GitHubException:
            # querying the environments might fail for private repos, ignore exceptions
            return []

    async def update_environment(self, org_id: str, repo_name: str, env_name: str, env: dict[str, Any]) -> None:
        print_debug(f"updating environment '{env_name}' for repo '{org_id}/{repo_name}'")

        if "name" in env:
            env.pop("name")

        if "branch_policies" in env:
            branch_policies = env.pop("branch_policies")
        else:
            branch_policies = None

        try:
            await self.requester.async_request_json("PUT", f"/repos/{org_id}/{repo_name}/environments/{env_name}", env)

            if branch_policies is not None:
                await self._update_deployment_branch_policies(org_id, repo_name, env_name, branch_policies)

            print_debug(f"updated repo environment '{env_name}'")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed to update repo environment '{env_name}':\n{ex}").with_traceback(tb)

    async def add_environment(self, org_id: str, repo_name: str, env_name: str, data: dict[str, Any]) -> None:
        print_debug(f"adding environment '{env_name}' for repo '{org_id}/{repo_name}'")
        await self.update_environment(org_id, repo_name, env_name, data)
        print_debug(f"added environment '{env_name}'")

    async def delete_environment(self, org_id: str, repo_name: str, env_name: str) -> None:
        print_debug(f"deleting repo environment '{env_name} for repo '{org_id}/{repo_name}'")

        status, _ = await self.requester.async_request_raw(
            "DELETE", f"/repos/{org_id}/{repo_name}/environments/{env_name}"
        )

        if status != 204:
            raise RuntimeError(f"failed to delete repo environment '{env_name}'")

        print_debug(f"removed repo environment '{env_name}'")

    async def _get_deployment_branch_policies(self, org_id: str, repo_name: str, env_name: str) -> list[dict[str, Any]]:
        print_debug(f"retrieving deployment branch policies for env '{env_name}'")

        try:
            url = f"/repos/{org_id}/{repo_name}/environments/{env_name}/deployment-branch-policies"
            response = await self.requester.async_request_json("GET", url)
            return response["branch_policies"]
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving deployment branch policies:\n{ex}").with_traceback(tb)

    async def _update_deployment_branch_policies(
        self, org_id: str, repo_name: str, env_name: str, branch_policies: list[str]
    ) -> None:
        print_debug(f"updating deployment branch policies for env '{env_name}'")

        try:
            current_branch_policies_by_name = associate_by_key(
                await self._get_deployment_branch_policies(org_id, repo_name, env_name),
                lambda x: x["name"],
            )
        except RuntimeError:
            current_branch_policies_by_name = {}

        try:
            for policy in branch_policies:
                if policy in current_branch_policies_by_name:
                    current_branch_policies_by_name.pop(policy)
                else:
                    await self._create_deployment_branch_policy(org_id, repo_name, env_name, policy)

            for policy_name, policy_dict in current_branch_policies_by_name.items():
                await self._delete_deployment_branch_policy(org_id, repo_name, env_name, policy_dict["id"])

            print_debug(f"updated deployment branch policies for env '{env_name}'")

        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed creating deployment branch policies:\n{ex}").with_traceback(tb)

    async def _create_deployment_branch_policy(self, org_id: str, repo_name: str, env_name: str, name: str) -> None:
        print_debug(f"creating deployment branch policy for env '{env_name}' with name '{name}")

        try:
            data = {"name": name}
            url = f"/repos/{org_id}/{repo_name}/environments/{env_name}/deployment-branch-policies"
            await self.requester.async_request_json("POST", url, data)
            print_debug(f"created deployment branch policy for env '{env_name}'")
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed creating deployment branch policy:\n{ex}").with_traceback(tb)

    async def _delete_deployment_branch_policy(
        self, org_id: str, repo_name: str, env_name: str, policy_id: int
    ) -> None:
        print_debug(f"deleting deployment branch policy for env '{env_name}' with id '{policy_id}")

        url = f"/repos/{org_id}/{repo_name}/environments/{env_name}/deployment-branch-policies/{policy_id}"
        status, body = await self.requester.async_request_raw("DELETE", url)

        if status != 204:
            raise RuntimeError(f"failed deleting deployment branch policy" f"\n{status}: {body}")

        print_debug(f"deleted deployment branch policy for env '{env_name}'")

    async def get_secrets(self, org_id: str, repo_name: str) -> list[dict[str, Any]]:
        print_debug(f"retrieving secrets for repo '{org_id}/{repo_name}'")

        try:
            status, body = await self.requester.async_request_raw("GET", f"/repos/{org_id}/{repo_name}/actions/secrets")
            if status == 200:
                return json.loads(body)["secrets"]
            else:
                return []
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving secrets for repo '{org_id}/{repo_name}':\n{ex}").with_traceback(tb)

    async def update_secret(self, org_id: str, repo_name: str, secret_name: str, secret: dict[str, Any]) -> None:
        print_debug(f"updating repo secret '{secret_name}' for repo '{org_id}/{repo_name}'")

        if "name" in secret:
            secret.pop("name")

        await self._encrypt_secret_inplace(org_id, repo_name, secret)

        status, _ = await self.requester.async_request_raw(
            "PUT",
            f"/repos/{org_id}/{repo_name}/actions/secrets/{secret_name}",
            json.dumps(secret),
        )

        if status != 204:
            raise RuntimeError(f"failed to update repo secret '{secret_name}'")

        print_debug(f"updated repo secret '{secret_name}'")

    async def add_secret(self, org_id: str, repo_name: str, data: dict[str, str]) -> None:
        secret_name = data.pop("name")
        print_debug(f"adding repo secret '{secret_name}' for repo '{org_id}/{repo_name}'")

        await self._encrypt_secret_inplace(org_id, repo_name, data)

        status, _ = await self.requester.async_request_raw(
            "PUT",
            f"/repos/{org_id}/{repo_name}/actions/secrets/{secret_name}",
            json.dumps(data),
        )

        if status != 201:
            raise RuntimeError(f"failed to add repo secret '{secret_name}'")

        print_debug(f"added repo secret '{secret_name}'")

    async def _encrypt_secret_inplace(self, org_id: str, repo_name: str, data: dict[str, Any]) -> None:
        value = data.pop("value")
        key_id, public_key = await self.get_public_key(org_id, repo_name)
        data["encrypted_value"] = encrypt_value(public_key, value)
        data["key_id"] = key_id

    async def delete_secret(self, org_id: str, repo_name: str, secret_name: str) -> None:
        print_debug(f"deleting repo secret '{secret_name}' for repo '{org_id}/{repo_name}'")

        status, _ = await self.requester.async_request_raw(
            "DELETE", f"/repos/{org_id}/{repo_name}/actions/secrets/{secret_name}"
        )

        if status != 204:
            raise RuntimeError(f"failed to delete repo secret '{secret_name}'")

        print_debug(f"removed repo secret '{secret_name}'")

    async def get_variables(self, org_id: str, repo_name: str) -> list[dict[str, Any]]:
        print_debug(f"retrieving variables for repo '{org_id}/{repo_name}'")

        try:
            status, body = await self.requester.async_request_raw(
                "GET", f"/repos/{org_id}/{repo_name}/actions/variables"
            )
            if status == 200:
                return json.loads(body)["variables"]
            else:
                return []
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving variables for repo '{org_id}/{repo_name}':\n{ex}").with_traceback(tb)

    async def update_variable(self, org_id: str, repo_name: str, variable_name: str, variable: dict[str, Any]) -> None:
        print_debug(f"updating repo variable '{variable_name}' for repo '{org_id}/{repo_name}'")

        if "name" in variable:
            variable.pop("name")

        status, body = await self.requester.async_request_raw(
            "PATCH",
            f"/repos/{org_id}/{repo_name}/actions/variables/{variable_name}",
            json.dumps(variable),
        )
        if status != 204:
            raise RuntimeError(f"failed to update repo variable '{variable_name}': {body}")

        print_debug(f"updated repo variable '{variable_name}'")

    async def add_variable(self, org_id: str, repo_name: str, data: dict[str, str]) -> None:
        variable_name = data.get("name")
        print_debug(f"adding repo variable '{variable_name}' for repo '{org_id}/{repo_name}'")

        status, body = await self.requester.async_request_raw(
            "POST",
            f"/repos/{org_id}/{repo_name}/actions/variables",
            json.dumps(data),
        )

        if status != 201:
            raise RuntimeError(f"failed to add repo variable '{variable_name}': {body}")

        print_debug(f"added repo variable '{variable_name}'")

    async def delete_variable(self, org_id: str, repo_name: str, variable_name: str) -> None:
        print_debug(f"deleting repo variable '{variable_name}' for repo '{org_id}/{repo_name}'")

        status, _ = await self.requester.async_request_raw(
            "DELETE", f"/repos/{org_id}/{repo_name}/actions/variables/{variable_name}"
        )

        if status != 204:
            raise RuntimeError(f"failed to delete repo variable '{variable_name}'")

        print_debug(f"removed repo variable '{variable_name}'")

    async def get_workflow_settings(self, org_id: str, repo_name: str) -> dict[str, Any]:
        print_debug(f"retrieving workflow settings for repo '{org_id}/{repo_name}'")

        workflow_settings: dict[str, Any] = {}

        try:
            permissions = await self.requester.async_request_json(
                "GET", f"/repos/{org_id}/{repo_name}/actions/permissions"
            )
            workflow_settings.update(permissions)
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(
                f"failed retrieving workflow settings for repo '{org_id}/{repo_name}':\n{ex}"
            ).with_traceback(tb)

        allowed_actions = permissions.get("allowed_actions", "none")
        if allowed_actions == "selected":
            workflow_settings.update(await self._get_selected_actions_for_workflow_settings(org_id, repo_name))

        if permissions.get("enabled", False) is not False:
            workflow_settings.update(await self._get_default_workflow_permissions(org_id, repo_name))

        return workflow_settings

    async def update_workflow_settings(self, org_id: str, repo_name: str, data: dict[str, Any]) -> None:
        print_debug(f"updating workflow settings for repo '{org_id}/{repo_name}'")

        permission_data = {k: data[k] for k in ["enabled", "allowed_actions"] if k in data}
        if len(permission_data) > 0:
            status, body = await self.requester.async_request_raw(
                "PUT", f"/repos/{org_id}/{repo_name}/actions/permissions", json.dumps(permission_data)
            )

            if status != 204:
                raise RuntimeError(
                    f"failed to update workflow settings for repo '{org_id}/{repo_name}'" f"\n{status}: {body}"
                )

            print_debug(f"updated workflow settings for repo '{org_id}/{repo_name}'")

        # only update the selected actions if needed
        if data.get("allowed_actions", "selected") == "selected":
            allowed_action_data = {
                k: data[k] for k in ["github_owned_allowed", "verified_allowed", "patterns_allowed"] if k in data
            }
            if len(allowed_action_data) > 0:
                await self._update_selected_actions_for_workflow_settings(org_id, repo_name, allowed_action_data)

        default_permission_data = {
            k: data[k] for k in ["default_workflow_permissions", "can_approve_pull_request_reviews"] if k in data
        }
        if len(default_permission_data) > 0:
            await self._update_default_workflow_permissions(org_id, repo_name, default_permission_data)

        print_debug(f"updated {len(data)} workflow setting(s)")

    async def _get_selected_actions_for_workflow_settings(self, org_id: str, repo_name: str) -> dict[str, Any]:
        print_debug(f"retrieving allowed actions for org '{org_id}'")

        try:
            return await self.requester.async_request_json(
                "GET", f"/repos/{org_id}/{repo_name}/actions/permissions/selected-actions"
            )
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(
                f"failed retrieving allowed actions for repo '{org_id}/{repo_name}':\n{ex}"
            ).with_traceback(tb)

    async def _update_selected_actions_for_workflow_settings(
        self, org_id: str, repo_name: str, data: dict[str, Any]
    ) -> None:
        print_debug(f"updating allowed actions for repo '{org_id}/{repo_name}'")

        status, body = await self.requester.async_request_raw(
            "PUT", f"/repos/{org_id}/{repo_name}/actions/permissions/selected-actions", json.dumps(data)
        )

        if status != 204:
            raise RuntimeError(f"failed updating allowed actions for repo '{org_id}/{repo_name}'" f"\n{status}: {body}")

        print_debug(f"updated allowed actions for repo '{org_id}/{repo_name}'")

    async def _get_default_workflow_permissions(self, org_id: str, repo_name: str) -> dict[str, Any]:
        print_debug(f"async retrieving default workflow permissions for repo '{org_id}/{repo_name}'")

        try:
            return await self.requester.async_request_json(
                "GET", f"/repos/{org_id}/{repo_name}/actions/permissions/workflow"
            )
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(
                f"failed retrieving default workflow permissions for repo '{org_id}/{repo_name}':\n{ex}"
            ).with_traceback(tb)

    async def _update_default_workflow_permissions(self, org_id: str, repo_name: str, data: dict[str, Any]) -> None:
        print_debug(f"updating default workflow permissions for repo '{org_id}/{repo_name}'")

        status, body = await self.requester.async_request_raw(
            "PUT", f"/repos/{org_id}/{repo_name}/actions/permissions/workflow", json.dumps(data)
        )

        if status != 204:
            raise RuntimeError(
                f"failed updating default workflow permissions for repo '{org_id}/{repo_name}'" f"\n{status}: {body}"
            )

        print_debug(f"updated default workflow permissions for repo '{org_id}/{repo_name}'")

    async def get_public_key(self, org_id: str, repo_name: str) -> tuple[str, str]:
        print_debug(f"retrieving repo public key for repo '{org_id}/{repo_name}'")

        try:
            response = await self.requester.async_request_json(
                "GET", f"/repos/{org_id}/{repo_name}/actions/secrets/public-key"
            )
            return response["key_id"], response["key"]
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving repo public key:\n{ex}").with_traceback(tb)

    async def dispatch_workflow(self, org_id: str, repo_name: str, workflow_name: str) -> bool:
        print_debug(f"dispatching workflow for repo '{org_id}/{repo_name}'")

        repo_data = await self.get_repo_data(org_id, repo_name)
        data = {"ref": repo_data["default_branch"]}

        status, _ = await self.requester.async_request_raw(
            "POST", f"/repos/{org_id}/{repo_name}/actions/workflows/{workflow_name}/dispatches", json.dumps(data)
        )

        if status != 204:
            print_debug(f"failed dispatching workflow for repo '{org_id}/{repo_name}'")
            return False
        else:
            print_debug(f"dispatched workflow for repo '{org_id}/{repo_name}'")
            return True

    async def get_ref_for_pull_request(self, org_id: str, repo_name: str, pull_number: str) -> str:
        print_debug(f"retrieving ref for pull request {pull_number} at {org_id}/{repo_name}")

        try:
            response = await self.requester.async_request_json(
                "GET", f"/repos/{org_id}/{repo_name}/pulls/{pull_number}"
            )
            return response["head"]["sha"]
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving ref for pull request:\n{ex}").with_traceback(tb)

    async def sync_from_template_repository(
        self,
        org_id: str,
        repo_name: str,
        template_repository: str,
        template_paths: Optional[list[str]],
    ) -> list[str]:
        template_owner, template_repo = re.split("/", template_repository, 1)

        updated_files = []
        with tempfile.TemporaryDirectory() as tmp_dir:
            archive_file_name = os.path.join(tmp_dir, "archive.zip")
            with open(archive_file_name, "wb") as archive_file:
                await self._download_repository_archive(archive_file, template_owner, template_repo)

            archive_target_dir = os.path.join(tmp_dir, "contents")
            with zipfile.ZipFile(archive_file_name, "r") as zip_file:
                zip_file.extractall(archive_target_dir)

            template_paths_set = set(template_paths) if isinstance(template_paths, list) else set()

            base_dir = None
            for path in pathlib.Path(archive_target_dir).rglob("*"):
                # the downloaded archive starts with a subdir that encodes
                # the name / hash of the downloaded repo, use that as the base dir
                # to resolve relative path names for updating the content.
                if base_dir is None:
                    base_dir = path

                relative_path = path.relative_to(base_dir)

                if path.is_file():
                    print_debug(f"updating file {relative_path}")

                    with open(path, "r") as file:
                        content = file.read()

                        if str(relative_path) in template_paths_set:
                            content = self._render_template_content(org_id, repo_name, content)

                        updated = await self.rest_api.content.update_content(
                            org_id, repo_name, str(relative_path), content
                        )
                        if updated:
                            updated_files.append(str(relative_path))

        return updated_files

    async def _download_repository_archive(self, file: IO, org_id: str, repo_name: str, ref: str = "") -> None:
        print_debug(f"downloading repository archive for '{org_id}/{repo_name}'")

        try:
            # TODO: use async streaming
            with self.requester.request_raw("GET", f"/repos/{org_id}/{repo_name}/zipball/{ref}") as response:
                file.write(response.content)

        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(
                f"failed retrieving repository archive from " f"repo '{org_id}/{repo_name}':\n{ex}"
            ).with_traceback(tb)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import re
from logging import getLogger

from pydantic import ValidationError
from quart import Response, current_app

from otterdog.utils import LogLevel
from otterdog.webapp.tasks import get_otterdog_config, refresh_otterdog_config
from otterdog.webapp.tasks.apply_changes import ApplyChangesTask
from otterdog.webapp.tasks.check_sync import CheckConfigurationInSyncTask
from otterdog.webapp.tasks.help_comment import HelpCommentTask
from otterdog.webapp.tasks.retrieve_team_membership import RetrieveTeamMembershipTask
from otterdog.webapp.tasks.validate_pull_request import ValidatePullRequestTask

from .github_models import IssueCommentEvent, PullRequestEvent, PushEvent
from .github_webhook import GitHubWebhook

webhook = GitHubWebhook()

logger = getLogger(__name__)


@webhook.hook("pull_request")
async def on_pull_request_received(data):
    try:
        event = PullRequestEvent.model_validate(data)
    except ValidationError:
        logger.error("failed to load pull request event data", exc_info=True)
        return success()

    if event.installation is None or event.organization is None:
        return success()

    otterdog_config = get_otterdog_config()

    if event.repository.name != otterdog_config.default_config_repo:
        return success()

    if event.action in ["opened", "ready_for_review"] and event.pull_request.draft is False:
        current_app.add_background_task(
            RetrieveTeamMembershipTask(
                event.installation.id,
                event.organization.login,
                event.repository,
                event.pull_request,
            ).execute
        )

    if event.action in ["opened", "synchronize", "ready_for_review", "reopened"] and event.pull_request.draft is False:
        current_app.add_background_task(
            ValidatePullRequestTask(
                event.installation.id,
                event.organization.login,
                event.repository,
                event.pull_request,
            ).execute
        )

    elif event.action in ["closed"] and event.pull_request.merged is True:
        current_app.add_background_task(
            ApplyChangesTask(
                event.installation.id,
                event.organization.login,
                event.repository,
                event.pull_request,
            ).execute
        )

    return success()


@webhook.hook("issue_comment")
async def on_issue_comment_received(data):
    try:
        event = IssueCommentEvent.model_validate(data)
    except ValidationError:
        logger.error("failed to load issue comment event data", exc_info=True)
        return success()

    if event.installation is None or event.organization is None:
        return success()

    otterdog_config = get_otterdog_config()

    if event.repository.name != otterdog_config.default_config_repo:
        return success()

    if event.action in ["created", "edited"] and "/pull/" in event.issue.html_url:
        org_id = event.organization.login
        installation_id = event.installation.id

        if re.match(r"\s*/help\s*", event.comment.body) is not None:
            current_app.add_background_task(
                HelpCommentTask(
                    installation_id,
                    org_id,
                    event.repository.name,
                    event.issue.number,
                ).execute
            )
            return success()
        elif re.match(r"\s*/team-info\s*", event.comment.body) is not None:
            current_app.add_background_task(
                RetrieveTeamMembershipTask(
                    installation_id,
                    org_id,
                    event.repository,
                    event.issue.number,
                ).execute
            )
            return success()
        elif re.match(r"\s*/check-sync\s*", event.comment.body) is not None:
            current_app.add_background_task(
                CheckConfigurationInSyncTask(
                    installation_id,
                    org_id,
                    event.repository,
                    event.issue.number,
                ).execute
            )
            return success()

        m = re.match(r"\s*/validate(\s+info)?\s*", event.comment.body)
        if m is None:
            return success()

        log_level_str = m.group(1)
        log_level = LogLevel.WARN

        if log_level_str is not None and log_level_str.strip() == "info":
            log_level = LogLevel.INFO

        current_app.add_background_task(
            ValidatePullRequestTask(
                installation_id,
                org_id,
                event.repository,
                event.issue.number,
                log_level,
            ).execute
        )

    return success()


@webhook.hook("push")
async def on_push_received(data):
    try:
        event = PushEvent.model_validate(data)
    except ValidationError:
        logger.error("failed to load push event data", exc_info=True)
        return success()

    if (
        event.repository.name != current_app.config["OTTERDOG_CONFIG_REPO"]
        or event.repository.owner.login != current_app.config["OTTERDOG_CONFIG_OWNER"]
    ):
        return success()

    if event.ref != f"refs/heads/{event.repository.default_branch}":
        return success()

    current_app.add_background_task(refresh_otterdog_config)
    return success()


def success() -> Response:
    return Response({}, mimetype="application/json", status=200)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
import os
from io import StringIO
from tempfile import TemporaryDirectory
from typing import Union, cast

from pydantic import ValidationError
from quart import current_app, render_template

from otterdog.config import OrganizationConfig
from otterdog.models import LivePatch
from otterdog.operations.diff_operation import DiffStatus
from otterdog.operations.plan import PlanOperation
from otterdog.providers.github import RestApi
from otterdog.utils import IndentingPrinter, LogLevel
from otterdog.webapp.tasks import Task, get_otterdog_config
from otterdog.webapp.tasks.validate_pull_request import (
    escape_for_github,
    get_admin_team,
    get_config,
)
from otterdog.webapp.webhook.github_models import PullRequest, Repository


@dataclasses.dataclass(repr=False)
class CheckConfigurationInSyncTask(Task[bool]):
    """Checks whether the base ref is in sync with live settings."""

    installation_id: int
    org_id: str
    repository: Repository
    pull_request_or_number: PullRequest | int

    async def _pre_execute(self) -> None:
        rest_api = await self.get_rest_api(self.installation_id)

        if isinstance(self.pull_request_or_number, int):
            response = await rest_api.pull_request.get_pull_request(
                self.org_id, self.repository.name, str(self.pull_request_or_number)
            )
            try:
                self.pull_request = PullRequest.model_validate(response)
            except ValidationError as ex:
                self.logger.exception("failed to load pull request event data", exc_info=ex)
                return
        else:
            self.pull_request = cast(PullRequest, self.pull_request_or_number)

        self.logger.info(
            "checking if base ref is in sync for pull request #%d of repo '%s'",
            self.pull_request.number,
            self.repository.full_name,
        )

        await self._create_pending_status(rest_api)

    async def _post_execute(self, result_or_exception: Union[bool, Exception]) -> None:
        rest_api = await self.get_rest_api(self.installation_id)

        if isinstance(result_or_exception, Exception):
            await self._create_failure_status(rest_api)
        else:
            await self._update_final_status(rest_api, result_or_exception)

    async def _execute(self) -> bool:
        otterdog_config = get_otterdog_config()
        pull_request_number = str(self.pull_request.number)
        project_name = otterdog_config.get_project_name(self.org_id) or self.org_id

        rest_api = await self.get_rest_api(self.installation_id)

        with TemporaryDirectory(dir=otterdog_config.jsonnet_base_dir) as work_dir:
            org_config = OrganizationConfig.of(
                project_name,
                self.org_id,
                {"provider": "inmemory", "api_token": rest_api.token},
                work_dir,
                otterdog_config,
            )

            jsonnet_config = org_config.jsonnet_config
            if not os.path.exists(jsonnet_config.org_dir):
                os.makedirs(jsonnet_config.org_dir)

            jsonnet_config.init_template()

            # get BASE config
            base_file = jsonnet_config.org_config_file
            await get_config(
                rest_api,
                self.org_id,
                self.org_id,
                otterdog_config.default_config_repo,
                base_file,
                self.pull_request.base.ref,
            )

            output = StringIO()
            printer = IndentingPrinter(output, log_level=LogLevel.ERROR)
            operation = PlanOperation(True, False, False, "")

            config_in_sync = True

            def sync_callback(org_id: str, diff_status: DiffStatus, patches: list[LivePatch]):
                nonlocal config_in_sync
                config_in_sync = diff_status.total_changes(True) == 0

            operation.set_callback(sync_callback)
            operation.init(otterdog_config, printer)

            await operation.execute(org_config)

            sync_output = output.getvalue()
            self.logger.info("sync plan: " + sync_output)

            if config_in_sync is False:
                comment = await render_template(
                    "out_of_sync_comment.txt",
                    result=escape_for_github(sync_output),
                    admin_team=f"{self.org_id}/{get_admin_team()}",
                )
            else:
                comment = await render_template("in_sync_comment.txt")

            await rest_api.issue.create_comment(
                self.org_id, otterdog_config.default_config_repo, pull_request_number, comment
            )

            return config_in_sync

    async def _create_pending_status(self, rest_api: RestApi):
        await rest_api.commit.create_commit_status(
            self.org_id,
            self.repository.name,
            self.pull_request.head.sha,
            "pending",
            _get_webhook_sync_context(),
            "checking if configuration is in-sync using otterdog",
        )

    async def _create_failure_status(self, rest_api: RestApi):
        await rest_api.commit.create_commit_status(
            self.org_id,
            self.repository.name,
            self.pull_request.head.sha,
            "failure",
            _get_webhook_sync_context(),
            "otterdog sync check failed, please contact an admin",
        )

    async def _update_final_status(self, rest_api: RestApi, config_in_sync: bool) -> None:
        if config_in_sync is True:
            desc = "otterdog sync check completed successfully"
            status = "success"
        else:
            desc = "otterdog sync check failed, check comment history"
            status = "error"

        await rest_api.commit.create_commit_status(
            self.org_id,
            self.repository.name,
            self.pull_request.head.sha,
            status,
            _get_webhook_sync_context(),
            desc,
        )

    def __repr__(self) -> str:
        pull_request_number = (
            self.pull_request_or_number
            if isinstance(self.pull_request_or_number, int)
            else self.pull_request_or_number.number
        )
        return f"CheckConfigurationInSyncTask(repo={self.repository.full_name}, pull_request={pull_request_number})"


def _get_webhook_sync_context() -> str:
    return current_app.config["GITHUB_WEBHOOK_SYNC_CONTEXT"]



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from abc import ABC, abstractmethod
from typing import Any, MutableMapping

from requests.auth import AuthBase


class AuthImpl(AuthBase):
    @abstractmethod
    def update_headers_with_authorization(self, headers: MutableMapping[str, Any]) -> None:
        ...


class AuthStrategy(ABC):
    @abstractmethod
    def get_auth(self) -> AuthImpl:
        ...


def app_auth(app_id: str, private_key: str) -> AuthStrategy:
    from .app import AppAuthStrategy

    return AppAuthStrategy(app_id, private_key)


def token_auth(github_token: str) -> AuthStrategy:
    from .token import TokenAuthStrategy

    return TokenAuthStrategy(github_token)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
from typing import Optional

from otterdog.jsonnet import JsonnetConfig
from otterdog.models import LivePatch, LivePatchType
from otterdog.models.variable import Variable
from otterdog.providers.github import GitHubProvider


@dataclasses.dataclass
class RepositoryVariable(Variable):
    """
    Represents a Variable defined on repo level.
    """

    @property
    def model_object_name(self) -> str:
        return "repo_variable"

    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> Optional[str]:
        return f"orgs.{jsonnet_config.create_repo_variable}"

    @classmethod
    async def apply_live_patch(cls, patch: LivePatch, org_id: str, provider: GitHubProvider) -> None:
        from .repository import Repository

        match patch.patch_type:
            case LivePatchType.ADD:
                assert isinstance(patch.expected_object, RepositoryVariable)
                assert isinstance(patch.parent_object, Repository)
                await provider.add_repo_variable(
                    org_id,
                    patch.parent_object.name,
                    await patch.expected_object.to_provider_data(org_id, provider),
                )

            case LivePatchType.REMOVE:
                assert isinstance(patch.current_object, RepositoryVariable)
                assert isinstance(patch.parent_object, Repository)
                await provider.delete_repo_variable(org_id, patch.parent_object.name, patch.current_object.name)

            case LivePatchType.CHANGE:
                assert isinstance(patch.expected_object, RepositoryVariable)
                assert isinstance(patch.current_object, RepositoryVariable)
                assert isinstance(patch.parent_object, Repository)
                await provider.update_repo_variable(
                    org_id,
                    patch.parent_object.name,
                    patch.current_object.name,
                    await patch.expected_object.to_provider_data(org_id, provider),
                )



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from typing import Any

from otterdog.credentials import CredentialProvider, Credentials


class InmemoryVault(CredentialProvider):
    """
    A simple credential provider for storing tokens in memory.
    """

    KEY_API_TOKEN = "api_token"

    def get_credentials(self, org_name: str, data: dict[str, Any], only_token: bool = False) -> Credentials:
        if only_token is not True:
            raise RuntimeError("in-memory vault only contains github tokens")

        github_token = data[self.KEY_API_TOKEN]
        return Credentials(None, None, None, github_token)

    def get_secret(self, key_data: str) -> str:
        raise RuntimeError("in-memory vault does not support secrets")

    def __repr__(self):
        return "InmemoryProvider()"



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import asyncio
import dataclasses
import json
import os
from datetime import datetime
from io import StringIO
from typing import Any, Callable, Iterator, Optional

import jsonschema
from importlib_resources import as_file, files
from jsonbender import F, Forall, OptionalS, S, bend  # type: ignore

from otterdog import resources
from otterdog.config import JsonnetConfig, OtterdogConfig
from otterdog.models import (
    LivePatchContext,
    LivePatchHandler,
    ModelObject,
    PatchContext,
    ValidationContext,
)
from otterdog.models.branch_protection_rule import BranchProtectionRule
from otterdog.models.environment import Environment
from otterdog.models.organization_secret import OrganizationSecret
from otterdog.models.organization_settings import OrganizationSettings
from otterdog.models.organization_variable import OrganizationVariable
from otterdog.models.organization_webhook import OrganizationWebhook
from otterdog.models.organization_workflow_settings import OrganizationWorkflowSettings
from otterdog.models.repo_ruleset import RepositoryRuleset
from otterdog.models.repo_secret import RepositorySecret
from otterdog.models.repo_variable import RepositoryVariable
from otterdog.models.repo_webhook import RepositoryWebhook
from otterdog.models.repo_workflow_settings import RepositoryWorkflowSettings
from otterdog.models.repository import Repository
from otterdog.providers.github import GitHubProvider
from otterdog.utils import (
    IndentingPrinter,
    associate_by_key,
    is_debug_enabled,
    is_info_enabled,
    jsonnet_evaluate_file,
    print_debug,
)

_ORG_SCHEMA = json.loads(files(resources).joinpath("schemas/organization.json").read_text())


@dataclasses.dataclass
class GitHubOrganization:
    """
    Represents a GitHub Organization with its associated resources.
    """

    github_id: str
    settings: OrganizationSettings
    webhooks: list[OrganizationWebhook] = dataclasses.field(default_factory=list)
    secrets: list[OrganizationSecret] = dataclasses.field(default_factory=list)
    variables: list[OrganizationVariable] = dataclasses.field(default_factory=list)
    repositories: list[Repository] = dataclasses.field(default_factory=list)

    _secrets_resolved: bool = False

    @property
    def secrets_resolved(self) -> bool:
        return self._secrets_resolved

    def add_webhook(self, webhook: OrganizationWebhook) -> None:
        self.webhooks.append(webhook)

    def get_webhook(self, url: str) -> Optional[OrganizationWebhook]:
        return next(filter(lambda x: x.url == url, self.webhooks), None)  # type: ignore

    def set_webhooks(self, webhooks: list[OrganizationWebhook]) -> None:
        self.webhooks = webhooks

    def add_secret(self, secret: OrganizationSecret) -> None:
        self.secrets.append(secret)

    def get_secret(self, name: str) -> Optional[OrganizationSecret]:
        return next(filter(lambda x: x.name == name, self.secrets), None)  # type: ignore

    def set_secrets(self, secrets: list[OrganizationSecret]) -> None:
        self.secrets = secrets

    def add_variable(self, variable: OrganizationVariable) -> None:
        self.variables.append(variable)

    def get_variable(self, name: str) -> Optional[OrganizationVariable]:
        return next(filter(lambda x: x.name == name, self.variables), None)  # type: ignore

    def set_variables(self, variables: list[OrganizationVariable]) -> None:
        self.variables = variables

    def add_repository(self, repo: Repository) -> None:
        self.repositories.append(repo)

    def get_repository(self, repo_name: str) -> Optional[Repository]:
        return next(filter(lambda x: x.name == repo_name, self.repositories), None)  # type: ignore

    def set_repositories(self, repos: list[Repository]) -> None:
        self.repositories = repos

    def validate(self, template_dir: str) -> ValidationContext:
        context = ValidationContext(self, template_dir)
        self.settings.validate(context, self)

        for webhook in self.webhooks:
            webhook.validate(context, self)

        for secret in self.secrets:
            secret.validate(context, self)

        for repo in self.repositories:
            repo.validate(context, self)

        return context

    @staticmethod
    def _validate_org_config(data: dict[str, Any]) -> None:
        with as_file(files(resources).joinpath("schemas")) as resource_dir:
            schema_root = resource_dir.as_uri()
            resolver = jsonschema.validators.RefResolver(base_uri=f"{schema_root}/", referrer=data)
            jsonschema.validate(instance=data, schema=_ORG_SCHEMA, resolver=resolver)

    def get_model_objects(self) -> Iterator[tuple[ModelObject, Optional[ModelObject]]]:
        yield self.settings, None
        yield from self.settings.get_model_objects()

        for webhook in self.webhooks:
            yield webhook, None
            yield from webhook.get_model_objects()

        for secret in self.secrets:
            yield secret, None
            yield from secret.get_model_objects()

        for variable in self.variables:
            yield variable, None
            yield from variable.get_model_objects()

        for repo in self.repositories:
            yield repo, None
            yield from repo.get_model_objects()

    @classmethod
    def from_model_data(cls, data: dict[str, Any]) -> GitHubOrganization:
        # validate the input data with the json schema.
        cls._validate_org_config(data)

        mapping = {
            "github_id": S("github_id"),
            "settings": S("settings") >> F(lambda x: OrganizationSettings.from_model_data(x)),
            "webhooks": OptionalS("webhooks", default=[]) >> Forall(lambda x: OrganizationWebhook.from_model_data(x)),
            "secrets": OptionalS("secrets", default=[]) >> Forall(lambda x: OrganizationSecret.from_model_data(x)),
            "variables": OptionalS("variables", default=[])
            >> Forall(lambda x: OrganizationVariable.from_model_data(x)),
            "repositories": OptionalS("repositories", default=[]) >> Forall(lambda x: Repository.from_model_data(x)),
        }

        return cls(**bend(mapping, data))

    def resolve_secrets(self, secret_resolver: Callable[[str], str]) -> None:
        for webhook in self.webhooks:
            webhook.resolve_secrets(secret_resolver)

        for secret in self.secrets:
            secret.resolve_secrets(secret_resolver)

        for repo in self.repositories:
            repo.resolve_secrets(secret_resolver)

        self._secrets_resolved = True

    def copy_secrets(self, other_org: GitHubOrganization) -> None:
        for webhook in self.webhooks:
            other_webhook = other_org.get_webhook(webhook.url)
            if other_webhook is not None:
                webhook.copy_secrets(other_webhook)

        for secret in self.secrets:
            other_secret = other_org.get_secret(secret.name)
            if other_secret is not None:
                secret.copy_secrets(other_secret)

        for repo in self.repositories:
            other_repo = other_org.get_repository(repo.name)
            if other_repo is not None:
                repo.copy_secrets(other_repo)

    def to_jsonnet(self, config: JsonnetConfig, context: PatchContext) -> str:
        default_org = GitHubOrganization.from_model_data(config.default_org_config_for_org_id(self.github_id))

        output = StringIO()
        printer = IndentingPrinter(output)

        printer.println(f"local orgs = {config.import_statement};")
        printer.println()
        printer.println(f"orgs.{config.create_org}('{self.github_id}') {{")
        printer.level_up()

        # print organization settings
        printer.print("settings+:")
        self.settings.to_jsonnet(printer, config, context, False, default_org.settings)

        # print organization webhooks
        if len(self.webhooks) > 0:
            default_org_webhook = OrganizationWebhook.from_model_data(config.default_org_webhook_config)

            printer.println("webhooks+: [")
            printer.level_up()

            for webhook in self.webhooks:
                webhook.to_jsonnet(printer, config, context, False, default_org_webhook)

            printer.level_down()
            printer.println("],")

        # print organization secrets
        if len(self.secrets) > 0:
            default_org_secret = OrganizationSecret.from_model_data(config.default_org_secret_config)

            printer.println("secrets+: [")
            printer.level_up()

            for secret in self.secrets:
                secret.to_jsonnet(printer, config, context, False, default_org_secret)

            printer.level_down()
            printer.println("],")

        # print organization variables
        if len(self.variables) > 0:
            default_org_variable = OrganizationVariable.from_model_data(config.default_org_variable_config)

            printer.println("variables+: [")
            printer.level_up()

            for variable in self.variables:
                variable.to_jsonnet(printer, config, context, False, default_org_variable)

            printer.level_down()
            printer.println("],")

        # print repositories
        if len(self.repositories) > 0:
            repos_by_name = associate_by_key(self.repositories, lambda x: x.name)
            default_repos_by_name = associate_by_key(default_org.repositories, lambda x: x.name)

            # add all default repos which are not yet contained in repos
            for default_repo_name, default_repo in default_repos_by_name.items():
                if repos_by_name.get(default_repo_name) is None:
                    repos_by_name[default_repo_name] = default_repo

            default_org_repo = Repository.from_model_data(config.default_repo_config)

            printer.println("_repositories+:: [")
            printer.level_up()

            for repo_name, repo in sorted(repos_by_name.items()):
                if repo_name in default_repos_by_name:
                    other_repo = default_repos_by_name[repo_name]
                    extend = True
                else:
                    other_repo = default_org_repo
                    extend = False

                repo.to_jsonnet(printer, config, context, extend, other_repo)

            printer.level_down()
            printer.println("],")

        printer.level_down()
        printer.println("}")

        return output.getvalue()

    def generate_live_patch(
        self, current_organization: GitHubOrganization, context: LivePatchContext, handler: LivePatchHandler
    ) -> None:
        OrganizationSettings.generate_live_patch(self.settings, current_organization.settings, None, context, handler)

        OrganizationWebhook.generate_live_patch_of_list(
            self.webhooks,
            current_organization.webhooks,
            None,
            context,
            handler,
        )

        OrganizationSecret.generate_live_patch_of_list(
            self.secrets, current_organization.secrets, None, context, handler
        )

        OrganizationVariable.generate_live_patch_of_list(
            self.variables, current_organization.variables, None, context, handler
        )

        Repository.generate_live_patch_of_list(
            self.repositories, current_organization.repositories, None, context, handler
        )

    @classmethod
    def load_from_file(
        cls,
        github_id: str,
        config_file: str,
        config: OtterdogConfig,
        resolve_secrets: bool = False,
    ) -> GitHubOrganization:
        if not os.path.exists(config_file):
            msg = f"configuration file '{config_file}' for organization '{github_id}' does not exist"
            raise RuntimeError(msg)

        print_debug(f"loading configuration for organization {github_id} from file {config_file}")
        data = jsonnet_evaluate_file(config_file)

        org = cls.from_model_data(data)

        if resolve_secrets:
            org.resolve_secrets(config.get_secret)

        return org

    @classmethod
    async def load_from_provider(
        cls,
        github_id: str,
        jsonnet_config: JsonnetConfig,
        provider: GitHubProvider,
        no_web_ui: bool = False,
        printer: Optional[IndentingPrinter] = None,
    ) -> GitHubOrganization:
        start = datetime.now()
        if printer is not None and is_info_enabled():
            printer.println("\norganization settings: Reading...")

        # FIXME: this uses the keys from the model schema which might be different to the provider schema
        #        for now this is the same for organization settings, but there might be cases where it is different.
        default_settings = jsonnet_config.default_org_config["settings"]
        included_keys = set(default_settings.keys())
        github_settings = await provider.get_org_settings(github_id, included_keys, no_web_ui)

        if printer is not None and is_info_enabled():
            end = datetime.now()
            printer.println(f"organization settings: Read complete after {(end - start).total_seconds()}s")

        settings = OrganizationSettings.from_provider_data(github_id, github_settings)

        if "workflows" in included_keys:
            workflow_settings = await provider.get_org_workflow_settings(github_id)
            settings.workflows = OrganizationWorkflowSettings.from_provider_data(github_id, workflow_settings)

        org = cls(github_id, settings)

        if printer is not None and is_info_enabled():
            start = datetime.now()
            printer.println("\nwebhooks: Reading...")

        if jsonnet_config.default_org_webhook_config is not None:
            github_webhooks = await provider.get_org_webhooks(github_id)

            if printer is not None and is_info_enabled():
                end = datetime.now()
                printer.println(f"webhooks: Read complete after {(end - start).total_seconds()}s")

            for webhook in github_webhooks:
                org.add_webhook(OrganizationWebhook.from_provider_data(github_id, webhook))
        else:
            print_debug("not reading org webhooks, no default config available")

        if jsonnet_config.default_org_secret_config is not None:
            start = datetime.now()
            if printer is not None and is_info_enabled():
                printer.println("\nsecrets: Reading...")

            github_secrets = await provider.get_org_secrets(github_id)

            if printer is not None and is_info_enabled():
                end = datetime.now()
                printer.println(f"secrets: Read complete after {(end - start).total_seconds()}s")

            for secret in github_secrets:
                org.add_secret(OrganizationSecret.from_provider_data(github_id, secret))
        else:
            print_debug("not reading org secrets, no default config available")

        if jsonnet_config.default_org_variable_config is not None:
            start = datetime.now()
            if printer is not None and is_info_enabled():
                printer.println("\nvariables: Reading...")

            github_variables = await provider.get_org_variables(github_id)

            if printer is not None and is_info_enabled():
                end = datetime.now()
                printer.println(f"variables: Read complete after {(end - start).total_seconds()}s")

            for variable in github_variables:
                org.add_variable(OrganizationVariable.from_provider_data(github_id, variable))
        else:
            print_debug("not reading org secrets, no default config available")

        if jsonnet_config.default_repo_config is not None:
            for repo in await _load_repos_from_provider(github_id, provider, jsonnet_config, printer):
                org.add_repository(repo)
        else:
            print_debug("not reading repos, no default config available")

        return org


async def _process_single_repo(
    gh_client: GitHubProvider,
    github_id: str,
    repo_name: str,
    jsonnet_config: JsonnetConfig,
    teams: dict[str, Any],
    app_installations: dict[str, str],
) -> tuple[str, Repository]:
    rest_api = gh_client.rest_api

    # get repo data
    github_repo_data = await rest_api.repo.get_repo_data(github_id, repo_name)
    repo = Repository.from_provider_data(github_id, github_repo_data)

    github_repo_workflow_data = await rest_api.repo.get_workflow_settings(github_id, repo_name)
    repo.workflows = RepositoryWorkflowSettings.from_provider_data(github_id, github_repo_workflow_data)

    if jsonnet_config.default_branch_protection_rule_config is not None:
        # get branch protection rules of the repo
        rules = await gh_client.get_branch_protection_rules(github_id, repo_name)
        for github_rule in rules:
            repo.add_branch_protection_rule(BranchProtectionRule.from_provider_data(github_id, github_rule))
    else:
        print_debug("not reading branch protection rules, no default config available")

    # repository rulesets are not available for private repos and free plan
    # TODO: support rulesets in private repos with enterprise plan
    if jsonnet_config.default_repo_ruleset_config is not None and repo.private is False:
        # get rulesets of the repo
        rulesets = await rest_api.repo.get_rulesets(github_id, repo_name)
        for github_ruleset in rulesets:
            # FIXME: need to associate an app id to its slug
            #        GitHub does not support that atm, so we lookup the currently installed
            #        apps for an organization which provide a mapping from id to slug.
            for actor in github_ruleset.get("bypass_actors", []):
                if actor.get("actor_type", None) == "Integration":
                    actor_id = str(actor.get("actor_id", 0))
                    if actor_id in app_installations:
                        actor["app_slug"] = app_installations[actor_id]
                elif actor.get("actor_type", None) == "Team":
                    actor_id = str(actor.get("actor_id", 0))
                    if actor_id in teams:
                        actor["team_slug"] = teams[actor_id]

            for rule in github_ruleset.get("rules", []):
                if rule.get("type", None) == "required_status_checks":
                    required_status_checks = rule.get("parameters", {}).get("required_status_checks", [])
                    for status_check in required_status_checks:
                        integration_id = str(status_check.get("integration_id", 0))
                        if integration_id in app_installations:
                            status_check["app_slug"] = app_installations[integration_id]

            repo.add_ruleset(RepositoryRuleset.from_provider_data(github_id, github_ruleset))
    else:
        print_debug("not reading repo rulesets, no default config available")

    if jsonnet_config.default_org_webhook_config is not None:
        # get webhooks of the repo
        webhooks = await rest_api.repo.get_webhooks(github_id, repo_name)
        for github_webhook in webhooks:
            repo.add_webhook(RepositoryWebhook.from_provider_data(github_id, github_webhook))
    else:
        print_debug("not reading repo webhooks, no default config available")

    if jsonnet_config.default_repo_secret_config is not None:
        # get secrets of the repo
        secrets = await rest_api.repo.get_secrets(github_id, repo_name)
        for github_secret in secrets:
            repo.add_secret(RepositorySecret.from_provider_data(github_id, github_secret))
    else:
        print_debug("not reading repo secrets, no default config available")

    if jsonnet_config.default_repo_variable_config is not None:
        # get variables of the repo
        variables = await rest_api.repo.get_variables(github_id, repo_name)
        for github_variable in variables:
            repo.add_variable(RepositoryVariable.from_provider_data(github_id, github_variable))
    else:
        print_debug("not reading repo variables, no default config available")

    if jsonnet_config.default_environment_config is not None:
        # get environments of the repo
        environments = await rest_api.repo.get_environments(github_id, repo_name)
        for github_environment in environments:
            repo.add_environment(Environment.from_provider_data(github_id, github_environment))
    else:
        print_debug("not reading environments, no default config available")

    if is_debug_enabled():
        print_debug(f"done retrieving data for repo '{repo_name}'")

    return repo_name, repo


async def _load_repos_from_provider(
    github_id: str, provider: GitHubProvider, jsonnet_config: JsonnetConfig, printer: Optional[IndentingPrinter] = None
) -> list[Repository]:
    start = datetime.now()
    if printer is not None and is_info_enabled():
        printer.println("\nrepositories: Reading...")

    repo_names = await provider.get_repos(github_id)

    teams = {
        str(team["id"]): f"{github_id}/{team['slug']}" for team in await provider.rest_api.org.get_teams(github_id)
    }

    app_installations = {
        str(installation["app_id"]): installation["app_slug"]
        for installation in await provider.rest_api.org.get_app_installations(github_id)
    }

    result = await asyncio.gather(
        *[
            _process_single_repo(provider, github_id, repo_name, jsonnet_config, teams, app_installations)
            for repo_name in repo_names
        ]
    )

    github_repos = []
    for data in result:
        _, repo_data = data
        github_repos.append(repo_data)

    if printer is not None and is_info_enabled():
        end = datetime.now()
        printer.println(f"repositories: Read complete after {(end - start).total_seconds()}s")

    return github_repos



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import os
import subprocess
from typing import Any

from otterdog import utils
from otterdog.credentials import CredentialProvider, Credentials


class PassVault(CredentialProvider):
    """
    A class to provide convenient access to a pass vault.
    """

    KEY_API_TOKEN = "api_token"
    KEY_USERNAME = "username"
    KEY_PASSWORD = "password"
    KEY_2FA_SEED = "2fa_seed"

    def __init__(
        self,
        password_store_dir: str,
        username_pattern: str,
        password_pattern: str,
        twofa_seed_pattern: str,
        api_token_pattern: str,
    ):
        utils.print_debug("accessing pass vault")
        status, output = subprocess.getstatusoutput("pass ls")
        if status != 0:
            raise RuntimeError(f"could not access pass vault:\n{output}")

        if password_store_dir:
            utils.print_debug(f"setting password store dir to '{password_store_dir}'")
            os.environ["PASSWORD_STORE_DIR"] = password_store_dir

        self._username_pattern = username_pattern
        self._password_pattern = password_pattern
        self._twofa_seed_pattern = twofa_seed_pattern
        self._api_token_pattern = api_token_pattern

        if status > 0:
            raise RuntimeError("pass vault is not accessible")

    def get_credentials(self, org_name: str, data: dict[str, Any], only_token: bool = False) -> Credentials:
        github_token = self._retrieve_key(self.KEY_API_TOKEN, org_name, data)

        if only_token is False:
            username = self._retrieve_key(self.KEY_USERNAME, org_name, data)
            password = self._retrieve_key(self.KEY_PASSWORD, org_name, data)
            totp_secret = self._retrieve_key(self.KEY_2FA_SEED, org_name, data)
        else:
            username = None
            password = None
            totp_secret = None

        return Credentials(username, password, totp_secret, github_token)

    def get_secret(self, key_data: str) -> str:
        return self._retrieve_resolved_key(key_data)

    def _retrieve_key(self, key: str, org_name: str, data: dict[str, str]) -> str:
        resolved_key = data.get(key)
        strict = True

        if resolved_key is None:
            match key:
                case PassVault.KEY_USERNAME:
                    pattern = self._username_pattern
                case PassVault.KEY_PASSWORD:
                    pattern = self._password_pattern
                case PassVault.KEY_2FA_SEED:
                    pattern = self._twofa_seed_pattern
                case PassVault.KEY_API_TOKEN:
                    pattern = self._api_token_pattern
                    strict = False
                case _:
                    raise RuntimeError(f"unexpected key '{key}'")

            if pattern:
                resolved_key = pattern.format(org_name)

        if resolved_key is None:
            raise RuntimeError(f"required key '{key}' not found in credential data")

        return PassVault._retrieve_resolved_key(resolved_key, strict)

    @staticmethod
    def _retrieve_resolved_key(key: str, strict: bool = True) -> str:
        status, secret = subprocess.getstatusoutput(f"pass {key} 2>/dev/null")
        if status != 0:
            # run the process again, capturing any error output for debugging.
            _, output = subprocess.getstatusoutput(f"pass {key}")

            if strict:
                raise RuntimeError(f"{key} could not be retrieved from your pass vault:\n{output}")
            else:
                utils.print_warn(f"{key} could not be retrieved from your pass vault:\n{output}")
                secret = ""

        return secret

    def __repr__(self):
        return "PassVault()"



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import abc
import dataclasses
from typing import Any, ClassVar, Optional

from jsonbender import OptionalS, S, bend  # type: ignore

from otterdog.jsonnet import JsonnetConfig
from otterdog.models import FailureType, ModelObject, PatchContext, ValidationContext
from otterdog.providers.github import GitHubProvider
from otterdog.utils import (
    UNSET,
    IndentingPrinter,
    is_set_and_valid,
    is_unset,
    write_patch_object_as_json,
)


@dataclasses.dataclass
class WorkflowSettings(ModelObject, abc.ABC):
    """
    Represents workflow settings on organizational / repository level.
    """

    allowed_actions: str
    allow_github_owned_actions: bool
    allow_verified_creator_actions: bool
    allow_action_patterns: list[str]
    default_workflow_permissions: str
    actions_can_approve_pull_request_reviews: bool

    _selected_action_properties: ClassVar[list[str]] = [
        "allow_github_owned_actions",
        "allow_verified_creator_actions",
        "allow_action_patterns",
    ]

    _allowed_actions_level: ClassVar[dict[str, int]] = {
        "all": 1,
        "local_only": 3,
        "selected": 2,
    }

    @classmethod
    def get_allowed_actions_level(cls, allowed_actions: str) -> int:
        if is_set_and_valid(allowed_actions):
            return cls._allowed_actions_level[allowed_actions]
        else:
            return 0

    def are_actions_more_restricted(self, repo_allowed_actions: str) -> bool:
        if is_set_and_valid(self.allowed_actions):
            org_level = WorkflowSettings.get_allowed_actions_level(self.allowed_actions)
            repo_level = WorkflowSettings.get_allowed_actions_level(repo_allowed_actions)
            return org_level >= repo_level
        return False

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        if is_set_and_valid(self.allowed_actions):
            if self.allowed_actions not in {"all", "local_only", "selected"}:
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'allowed_actions' of value "
                    f"'{self.allowed_actions}', "
                    f"only values ('all' | 'local_only' | 'selected') are allowed.",
                )

            if self.allowed_actions != "selected" and len(self.allow_action_patterns) > 0:
                context.add_failure(
                    FailureType.WARNING,
                    f"{self.get_model_header(parent_object)} has 'allowed_actions' set to "
                    f"'{self.allowed_actions}', "
                    f"but 'allow_action_patterns' is set to '{self.allow_action_patterns}', "
                    f"setting will be ignored.",
                )

        if is_set_and_valid(self.default_workflow_permissions):
            if self.default_workflow_permissions not in {"read", "write"}:
                context.add_failure(
                    FailureType.ERROR,
                    f"'default_workflow_permissions' has value '{self.default_workflow_permissions}', "
                    f"only values ('read' | 'write') are allowed.",
                )

    def include_field_for_diff_computation(self, field: dataclasses.Field) -> bool:
        if is_set_and_valid(self.allowed_actions) and self.allowed_actions in ["all", "local_only"]:
            if field.name in self._selected_action_properties:
                return False

        return True

    def include_field_for_patch_computation(self, field: dataclasses.Field) -> bool:
        if field.name in self._selected_action_properties:
            if self.__getattribute__(field.name) is None:
                return False

        return True

    @classmethod
    def from_model_data(cls, data: dict[str, Any]):
        mapping = {k: OptionalS(k, default=UNSET) for k in map(lambda x: x.name, cls.all_fields())}
        return cls(**bend(mapping, data))

    @classmethod
    def from_provider_data(cls, org_id: str, data: dict[str, Any]):
        mapping = cls.get_mapping_from_provider(org_id, data)
        return cls(**bend(mapping, data))

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = {k: OptionalS(k, default=UNSET) for k in map(lambda x: x.name, cls.all_fields())}
        mapping.update(
            {
                "allow_github_owned_actions": OptionalS("github_owned_allowed", default=None),
                "allow_verified_creator_actions": OptionalS("verified_allowed", default=None),
                "allow_action_patterns": OptionalS("patterns_allowed", default=[]),
                "actions_can_approve_pull_request_reviews": OptionalS(
                    "can_approve_pull_request_reviews", default=UNSET
                ),
            }
        )
        return mapping

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        mapping = {
            field.name: S(field.name) for field in cls.provider_fields() if not is_unset(data.get(field.name, UNSET))
        }

        if "allow_github_owned_actions" in data:
            mapping.pop("allow_github_owned_actions")
            mapping["github_owned_allowed"] = S("allow_github_owned_actions")

        if "allow_verified_creator_actions" in data:
            mapping.pop("allow_verified_creator_actions")
            mapping["verified_allowed"] = S("allow_verified_creator_actions")

        if "allow_action_patterns" in data:
            mapping.pop("allow_action_patterns")
            mapping["patterns_allowed"] = S("allow_action_patterns")

        if "actions_can_approve_pull_request_reviews" in data:
            mapping.pop("actions_can_approve_pull_request_reviews")
            mapping["can_approve_pull_request_reviews"] = S("actions_can_approve_pull_request_reviews")

        return mapping

    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> Optional[str]:
        return None

    def to_jsonnet(
        self,
        printer: IndentingPrinter,
        jsonnet_config: JsonnetConfig,
        context: PatchContext,
        extend: bool,
        default_object: ModelObject,
    ) -> None:
        patch = self.get_patch_to(default_object)
        write_patch_object_as_json(patch, printer, False)
        printer.level_down()
        printer.println("},")



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import os

from otterdog.config import OrganizationConfig
from otterdog.models.github_organization import GitHubOrganization
from otterdog.utils import sort_jsonnet, strip_trailing_commas, style

from ..models import PatchContext
from . import Operation


class CanonicalDiffOperation(Operation):
    """
    Performs a canonical diff for organization configurations, i.e. comparing the current configuration
    with a canonical version and showing the unified diff of it.
    """

    def __init__(self):
        super().__init__()

    def pre_execute(self) -> None:
        self.printer.println("Showing canonical diff:")

    async def execute(self, org_config: OrganizationConfig) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config
        jsonnet_config.init_template()

        self.printer.println(f"\nOrganization {style(org_config.name, bright=True)}[id={github_id}]")

        org_file_name = jsonnet_config.org_config_file

        if not os.path.exists(org_file_name):
            self.printer.print_error(
                f"configuration file '{org_file_name}' does not yet exist, run fetch-config or import first."
            )
            return 1

        try:
            organization = GitHubOrganization.load_from_file(github_id, org_file_name, self.config)
        except RuntimeError as ex:
            self.printer.print_error(f"failed to load configuration: {str(ex)}")
            return 1

        with open(org_file_name, "r") as file:
            original_config = file.read()

        original_config_without_comments: list[str] = strip_trailing_commas(
            sort_jsonnet(list(filter(lambda x: not x.strip().startswith("#"), original_config.split("\n"))))
        )

        context = PatchContext(github_id, organization.settings)
        canonical_config = organization.to_jsonnet(jsonnet_config, context)
        canonical_config_as_lines = strip_trailing_commas(sort_jsonnet(canonical_config.split("\n")))

        self.printer.println()

        for line in self._diff(canonical_config_as_lines, original_config_without_comments, "canonical", "original"):
            if line.startswith("+"):
                self.printer.println(style(line, fg="green"))
            elif line.startswith("-"):
                self.printer.println(style(line, fg="red"))
            else:
                self.printer.println(line)

        return 0

    @staticmethod
    def _diff(a, b, name_a, name_b):
        from subprocess import PIPE, Popen
        from tempfile import NamedTemporaryFile

        with NamedTemporaryFile() as file:
            file.write(bytes("\n".join(a), "utf-8"))
            file.flush()
            p = Popen(
                ["diff", "--label", name_a, "--label", name_b, "-u", "-w", "-", file.name], stdin=PIPE, stdout=PIPE
            )
            out, err = p.communicate(bytes("\n".join(b), "utf-8"))

            for line in out.decode("utf-8").split("\n"):
                yield line



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from abc import ABC
from datetime import datetime
from functools import cached_property
from typing import Optional

from otterdog.providers.github.auth import AuthStrategy

from .requester import Requester


class RestApi:
    # use a fixed API version
    _GH_API_VERSION = "2022-11-28"
    _GH_API_URL_ROOT = "https://api.github.com"

    def __init__(self, auth_strategy: AuthStrategy):
        self._auth_strategy = auth_strategy
        self._requester = Requester(auth_strategy, self._GH_API_URL_ROOT, self._GH_API_VERSION)

    @property
    def token(self) -> Optional[str]:
        from otterdog.providers.github.auth.token import TokenAuthStrategy

        if isinstance(self._auth_strategy, TokenAuthStrategy):
            return self._auth_strategy.token
        else:
            return None

    def close(self) -> None:
        self._requester.close()

    @property
    def requester(self) -> Requester:
        return self._requester

    @cached_property
    def app(self):
        from .app_client import AppClient

        return AppClient(self)

    @cached_property
    def commit(self):
        from .commit_client import CommitClient

        return CommitClient(self)

    @cached_property
    def content(self):
        from .content_client import ContentClient

        return ContentClient(self)

    @cached_property
    def issue(self):
        from .issue_client import IssueClient

        return IssueClient(self)

    @cached_property
    def pull_request(self):
        from .pull_request_client import PullRequestClient

        return PullRequestClient(self)

    @cached_property
    def repo(self):
        from .repo_client import RepoClient

        return RepoClient(self)

    @cached_property
    def org(self):
        from .org_client import OrgClient

        return OrgClient(self)

    @cached_property
    def user(self):
        from .user_client import UserClient

        return UserClient(self)

    @cached_property
    def team(self):
        from .team_client import TeamClient

        return TeamClient(self)


class RestClient(ABC):
    def __init__(self, rest_api: RestApi):
        self.__rest_api = rest_api

    @property
    def rest_api(self) -> RestApi:
        return self.__rest_api

    @property
    def requester(self) -> Requester:
        return self.__rest_api.requester


def encrypt_value(public_key: str, secret_value: str) -> str:
    """
    Encrypt a Unicode string using a public key.
    """
    from base64 import b64encode

    from nacl import encoding, public

    public_key_obj = public.PublicKey(public_key.encode("utf-8"), encoding.Base64Encoder)
    sealed_box = public.SealedBox(public_key_obj)
    encrypted = sealed_box.encrypt(secret_value.encode("utf-8"))
    return b64encode(encrypted).decode("utf-8")


_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


def parse_date_string(date: str) -> datetime:
    return datetime.strptime(date, _FORMAT)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import os

from decouple import config  # type: ignore


class AppConfig(object):
    QUART_APP = "otterdog.webapp"

    APP_ROOT = config("APP_ROOT")
    if not os.path.exists(APP_ROOT):
        os.makedirs(APP_ROOT)

    OTTERDOG_CONFIG_URL = config("OTTERDOG_CONFIG_URL", default=None)
    OTTERDOG_CONFIG_OWNER = config("OTTERDOG_CONFIG_OWNER", default=None)
    OTTERDOG_CONFIG_REPO = config("OTTERDOG_CONFIG_REPO", default=None)

    # Set up the App SECRET_KEY
    SECRET_KEY = config("SECRET_KEY")

    GITHUB_ADMIN_TEAM = config("GITHUB_ADMIN_TEAM", default="otterdog-admins")
    GITHUB_WEBHOOK_ENDPOINT = config("GITHUB_WEBHOOK_ENDPOINT", default="/github-webhook/receive")
    GITHUB_WEBHOOK_SECRET = config("GITHUB_WEBHOOK_SECRET", default=None)
    GITHUB_WEBHOOK_VALIDATION_CONTEXT = config("GITHUB_WEBHOOK_VALIDATION_CONTEXT", default="otterdog-validate")
    GITHUB_WEBHOOK_SYNC_CONTEXT = config("GITHUB_WEBHOOK_SYNC_CONTEXT", default="otterdog-sync")

    # GitHub App config
    GITHUB_APP_ID = config("GITHUB_APP_ID")
    GITHUB_APP_PRIVATE_KEY = config("GITHUB_APP_PRIVATE_KEY")


class ProductionConfig(AppConfig):
    DEBUG = False

    # Security
    SESSION_COOKIE_HTTPONLY = True
    REMEMBER_COOKIE_HTTPONLY = True
    REMEMBER_COOKIE_DURATION = 3600


class DebugConfig(AppConfig):
    DEBUG = True


# Load all possible configurations
config_dict = {"Production": ProductionConfig, "Debug": DebugConfig}



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import base64
from typing import Any, Optional

from otterdog.providers.github.exception import GitHubException
from otterdog.utils import print_debug

from . import RestApi, RestClient


class ContentClient(RestClient):
    def __init__(self, rest_api: RestApi):
        super().__init__(rest_api)

    async def get_content_object(
        self, org_id: str, repo_name: str, path: str, ref: Optional[str] = None
    ) -> dict[str, Any]:
        print_debug(f"retrieving content '{path}' from repo '{org_id}/{repo_name}'")

        try:
            if ref is not None:
                params = {"ref": ref}
            else:
                params = None

            return await self.requester.async_request_json(
                "GET", f"/repos/{org_id}/{repo_name}/contents/{path}", params=params
            )
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving content '{path}' from repo '{repo_name}':\n{ex}").with_traceback(tb)

    async def get_content(self, org_id: str, repo_name: str, path: str, ref: Optional[str]) -> str:
        json_response = await self.get_content_object(org_id, repo_name, path, ref)
        return base64.b64decode(json_response["content"]).decode("utf-8")

    async def update_content(
        self,
        org_id: str,
        repo_name: str,
        path: str,
        content: str,
        message: Optional[str] = None,
    ) -> bool:
        print_debug(f"putting content '{path}' to repo '{org_id}/{repo_name}'")

        try:
            json_response = await self.get_content_object(org_id, repo_name, path)
            old_sha = json_response["sha"]
            old_content = base64.b64decode(json_response["content"]).decode("utf-8")
        except RuntimeError:
            old_sha = None
            old_content = None

        # check if the content has changed, otherwise do not update
        if old_content is not None and content == old_content:
            print_debug("not updating content, no changes")
            return False

        base64_encoded_data = base64.b64encode(content.encode("utf-8"))
        base64_content = base64_encoded_data.decode("utf-8")

        if message is None:
            push_message = f"Updating file '{path}' with otterdog."
        else:
            push_message = message

        data = {
            "message": push_message,
            "content": base64_content,
        }

        if old_sha is not None:
            data["sha"] = old_sha

        try:
            await self.requester.async_request_json("PUT", f"/repos/{org_id}/{repo_name}/contents/{path}", data)
            return True
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed putting content '{path}' to repo '{repo_name}':\n{ex}").with_traceback(tb)

    async def delete_content(
        self,
        org_id: str,
        repo_name: str,
        path: str,
        message: Optional[str] = None,
    ) -> bool:
        print_debug(f"deleting content '{path}' in repo '{org_id}/{repo_name}'")

        try:
            json_response = await self.get_content_object(org_id, repo_name, path)
            old_sha = json_response["sha"]
        except RuntimeError:
            old_sha = None

        if old_sha is None:
            return False

        if message is None:
            push_message = f"Deleting file '{path}' with otterdog."
        else:
            push_message = message

        data = {"message": push_message, "sha": old_sha}

        try:
            await self.requester.async_request_json("DELETE", f"/repos/{org_id}/{repo_name}/contents/{path}", data)
            return True
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed deleting content '{path}' in repo '{repo_name}':\n{ex}").with_traceback(tb)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from importlib import import_module
from importlib.util import find_spec

from quart import Quart

from .config import AppConfig

_BLUEPRINT_MODULES: list[str] = []


def register_github_webhook(app) -> None:
    webhook_fqn = "otterdog.webapp.webhook"
    spec = find_spec(webhook_fqn)
    if spec is not None:
        module = import_module(webhook_fqn)
        module.webhook.init_app(app)


def register_blueprints(app):
    for module_name in _BLUEPRINT_MODULES:
        routes_fqn = f"otterdog.webapp.{module_name}.routes"
        spec = find_spec(routes_fqn)
        if spec is not None:
            module = import_module(routes_fqn)
            app.register_blueprint(module.blueprint)


def create_app(app_config: AppConfig):
    app = Quart(app_config.QUART_APP)
    app.config.from_object(app_config)

    register_github_webhook(app)
    register_blueprints(app)

    return app



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from otterdog.utils import print_debug

from ..exception import GitHubException
from . import RestApi, RestClient


class UserClient(RestClient):
    def __init__(self, rest_api: RestApi):
        super().__init__(rest_api)

    async def get_user_ids(self, login: str) -> tuple[int, str]:
        print_debug(f"retrieving user ids for user '{login}'")

        try:
            response = await self.requester.async_request_json("GET", f"/users/{login}")
            return response["id"], response["node_id"]
        except GitHubException as ex:
            tb = ex.__traceback__
            raise RuntimeError(f"failed retrieving user node id:\n{ex}").with_traceback(tb)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import json
import re
import sys
from argparse import Namespace
from dataclasses import dataclass
from enum import Enum
from typing import (
    Any,
    Callable,
    Generic,
    Literal,
    Optional,
    Sequence,
    TextIO,
    Tuple,
    TypeGuard,
    TypeVar,
    Union,
)
from urllib.parse import urlparse

import click
from colorama import Style

T = TypeVar("T")


# verbose levels
# 0: off
# 1: info
# 2: debug
# 3: trace
_verbose_level = 0


def init(verbose: int) -> None:
    global _verbose_level
    _verbose_level = verbose


def is_info_enabled() -> bool:
    return _verbose_level >= 1


def is_debug_enabled() -> bool:
    return _verbose_level >= 2


def is_trace_enabled() -> bool:
    return _verbose_level >= 3


def style(
    text: str,
    fg: Optional[Union[int, Tuple[int, int, int], str]] = None,
    bg: Optional[Union[int, Tuple[int, int, int], str]] = None,
    bold: Optional[bool] = None,
    bright: Optional[bool] = None,
    dim: Optional[bool] = None,
    underline: Optional[bool] = None,
    overline: Optional[bool] = None,
    italic: Optional[bool] = None,
    blink: Optional[bool] = None,
    reverse: Optional[bool] = None,
    strikethrough: Optional[bool] = None,
    reset: bool = True,
) -> str:
    if bright is True:
        text = f"{Style.BRIGHT}{text}"
    return click.style(text, fg, bg, bold, dim, underline, overline, italic, blink, reverse, strikethrough, reset)


def print_info(msg: str, printer: TextIO = sys.stdout) -> None:
    if is_info_enabled():
        _print_message(msg, "green", "Info", printer)


def print_debug(msg: str, printer: TextIO = sys.stdout) -> None:
    if is_debug_enabled():
        printer.write(f"{style('[DEBUG]', fg='cyan')} {msg}\n")


def print_trace(msg: str, printer: TextIO = sys.stdout) -> None:
    if is_trace_enabled():
        printer.write(f"{style('[TRACE]', fg='magenta')} {msg}\n")


def print_warn(msg: str, printer: TextIO = sys.stdout) -> None:
    _print_message(msg, "yellow", "Warning", printer)


def print_error(msg: str, printer: TextIO = sys.stdout) -> None:
    _print_message(msg, "red", "Error", printer)


def _print_message(msg: str, color: str, level: str, printer: TextIO) -> None:
    printer.write(style("╷\n", fg=color))

    lines = msg.splitlines()
    level_prefix = style(f"│ {level}:", fg=color)

    if len(lines) > 1:
        printer.write(f"{level_prefix} {lines[0]}\n")
        printer.write(style("│\n", fg=color))
        for line in lines[1:]:
            printer.write(f"{style('│', fg=color)}    {line}\n")
    else:
        printer.write(f"{level_prefix} {msg}\n")

    printer.write(style("╵\n", fg=color))


class _Unset:
    """
    A marker class to indicate that a value is unset and thus should
    not be considered. This is different to None.
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(_Unset, cls).__new__(cls)
        return cls._instance

    def __repr__(self) -> str:
        return "<UNSET>"

    def __bool__(self) -> Literal[False]:
        return False

    def __copy__(self):
        return UNSET

    def __deepcopy__(self, memo: dict[int, Any]):
        return UNSET


UNSET = _Unset()


def is_unset(value: Any) -> bool:
    """
    Returns whether the given value is an instance of Unset.
    """
    return value is UNSET


def is_set_and_valid(value: Any) -> bool:
    return not is_unset(value) and value is not None


def is_set_and_present(value: Optional[T]) -> TypeGuard[T]:
    return is_set_and_valid(value)


@dataclass
class Change(Generic[T]):
    from_value: Optional[T]
    to_value: Optional[T]


def is_different_ignoring_order(value: Any, other_value: Any) -> bool:
    """
    Checks whether two values are considered to be equal.
    Note: two lists are considered to be equal if they contain the same elements,
    regardless or the order.
    """
    if isinstance(value, list):
        return sorted(value) != sorted(other_value)

    return value != other_value


def patch_to_other(value: Any, other_value: Any) -> tuple[bool, Any]:
    if isinstance(value, dict):
        raise ValueError("dictionary values not supported")
    elif isinstance(value, list):
        sorted_value_list = sorted(value)
        sorted_other_list = sorted(other_value)

        if sorted_value_list != sorted_other_list:
            diff = _diff_list(sorted_value_list, sorted_other_list)
            if len(diff) > 0:
                return True, diff
            else:
                return False, diff
    else:
        if value != other_value:
            return True, value

    # values are not different, no patch generated
    return False, None


def _diff_list(list1: list[T], list2: list[T]) -> list[T]:
    s = set(list2)
    return [x for x in list1 if x not in s]


def write_patch_object_as_json(
    diff_object: dict[str, Any], printer: IndentingPrinter, close_object: bool = True
) -> None:
    if close_object is True and len(diff_object) == 0:
        printer.println(",")
        return

    printer.println(" {")
    printer.level_up()

    for key, value in sorted(diff_object.items()):
        if is_unset(value):
            print_warn(f"key '{key}' defined in default configuration not present in config, skipping")
            continue

        if isinstance(value, list):
            printer.println(f"{key}+: [")
            printer.level_up()
            num_items = len(value)
            for index, item in enumerate(value):
                if index < num_items - 1:
                    printer.println(f"{json.dumps(item, ensure_ascii=False)},")
                else:
                    printer.println(f"{json.dumps(item, ensure_ascii=False)}")
            printer.level_down()
            printer.println("],")
        else:
            printer.println(f"{key}: {json.dumps(value, ensure_ascii=False)},")

    if close_object is True:
        printer.level_down()
        printer.println("},")


def associate_by_key(input_list: Sequence[T], key_func: Callable[[T], str]) -> dict[str, T]:
    result = {}
    for item in input_list:
        key = key_func(item)

        if key in result:
            raise RuntimeError(f"duplicate item found with key '{key}'")

        result[key] = item

    return result


def multi_associate_by_key(input_list: Sequence[T], key_func: Callable[[T], list[str]]) -> dict[str, T]:
    result = {}
    for item in input_list:
        keys = key_func(item)

        for key in keys:
            if key in result:
                raise RuntimeError(f"duplicate item found with key '{key}'")

            result[key] = item

    return result


class LogLevel(Enum):
    GLOBAL = 0
    INFO = 1
    WARN = 2
    ERROR = 3


class IndentingPrinter:
    def __init__(
        self, writer: TextIO, initial_offset: int = 0, spaces_per_level: int = 2, log_level: LogLevel = LogLevel.GLOBAL
    ):
        self._writer = writer
        self._initial_offset = " " * initial_offset
        self._level = 0
        self._spaces_per_level = spaces_per_level
        self._indented_line = False
        self._log_level = log_level

    @property
    def writer(self) -> TextIO:
        return self._writer

    @property
    def _current_indentation(self) -> str:
        return self._initial_offset + " " * (self._level * self._spaces_per_level)

    def print(self, text: str = "") -> None:
        lines = text.splitlines(keepends=True)
        if len(lines) > 0:
            for line in lines:
                self._print_indentation()

                if line.endswith("\n"):
                    self._writer.write(line[:-1])
                    self.print_line_break()
                else:
                    self._writer.write(line)

    def println(self, text: str = "") -> None:
        self.print(text)
        self.print_line_break()

    def print_line_break(self) -> None:
        self._writer.write("\n")
        self._indented_line = False

    def _print_indentation(self) -> None:
        if not self._indented_line:
            self._writer.write(self._current_indentation)
            self._indented_line = True

    def _is_logging_enabled(self, level: int, global_fn: Callable[[], bool]) -> bool:
        match self._log_level:
            case LogLevel.GLOBAL:
                return global_fn()

            case _:
                return self._log_level.value <= level

    def print_info(self, msg: str) -> None:
        if self._is_logging_enabled(1, is_info_enabled):
            _print_message(msg, "green", "Info", self._writer)

    def is_info_enabled(self) -> bool:
        return self._is_logging_enabled(1, is_info_enabled)

    def print_warn(self, msg: str) -> None:
        if self._is_logging_enabled(2, lambda: True):
            _print_message(msg, "yellow", "Warning", self._writer)

    def print_error(self, msg: str) -> None:
        if self._is_logging_enabled(3, lambda: True):
            _print_message(msg, "red", "Error", self._writer)

    def level_up(self) -> None:
        self._level += 1

    def level_down(self) -> None:
        self._level -= 1
        assert self._level >= 0


def jsonnet_evaluate_file(file: str) -> dict[str, Any]:
    import _gojsonnet  # type: ignore

    print_trace(f"evaluating jsonnet file {file}")

    try:
        return json.loads(_gojsonnet.evaluate_file(file))
    except Exception as ex:
        raise RuntimeError(f"failed to evaluate jsonnet file: {str(ex)}")


def jsonnet_evaluate_snippet(snippet: str) -> dict[str, Any]:
    import _gojsonnet  # type: ignore

    print_trace(f"evaluating jsonnet snippet {snippet}")

    try:
        return json.loads(_gojsonnet.evaluate_snippet("", snippet))
    except Exception as ex:
        raise RuntimeError(f"failed to evaluate snippet: {str(ex)}")


def get_or_default(namespace: Namespace, key: str, default: T) -> T:
    if namespace.__contains__(key):
        return namespace.__getattribute__(key)
    else:
        return default


def snake_to_camel_case(string: str) -> str:
    result = re.sub(r"[_\-]+", " ", string).title().replace(" ", "")
    return result[0].lower() + result[1:]


def camel_to_snake_case(string: str) -> str:
    string = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", string)
    string = re.sub("__([A-Z])", r"_\1", string)
    string = re.sub("([a-z0-9])([A-Z])", r"\1_\2", string)
    return string.lower()


def parse_template_url(url: str) -> tuple[str, str, str]:
    parts = urlparse(url)

    if parts.netloc != "github.com":
        raise ValueError(f"only github.com is supported for template urls: {parts.netloc}")

    repo_url = f"{parts.scheme}://{parts.netloc}{parts.path}"

    matcher = re.match(r"([^@]+)(@(.*))?", parts.fragment)
    if matcher is None:
        raise ValueError(f"failed to parse file from template url: {url}")

    file = matcher.group(1)
    ref = matcher.group(3)

    if ref is None:
        raise ValueError(f"failed to parse ref from template url: {url}")

    return repo_url, file, ref


def is_ghsa_repo(repo_name: str) -> bool:
    """
    Returns True if the given repo_name is considered to be a repo created for a GitHub security advisory.

    Pattern: <regular repo name>-ghsa-xxxx-xxxx-xxxx
    """
    return re.match(".*-ghsa(-[23456789cfghjmpqrvwx]{4}){3}", repo_name) is not None


def strip_trailing_commas(lines: list[str]) -> list[str]:
    return list(map(lambda line: line.rstrip(","), lines))


def sort_jsonnet(lines: list[str]) -> list[str]:
    ast: list[tuple[str, Any]] = []
    object_stack = [ast]

    for line in lines:
        trimmed_line = line.rstrip().rstrip(",")
        if trimmed_line.endswith("{") or trimmed_line.endswith("["):
            current_node: list[tuple[str, Any]] = []
            object_stack[-1].append((line, current_node))
            object_stack.append(current_node)
        elif trimmed_line.endswith("}") and "{" not in trimmed_line:
            object_stack[-1].append((line, None))
            object_stack.pop()
        elif trimmed_line.endswith("]") and "[" not in trimmed_line:
            object_stack[-1].append((line, None))
            object_stack.pop()
        else:
            object_stack[-1].append((line, None))

    for node in ast:
        _sort_node(node)

    result: list[str] = []
    for node in ast:
        print_node(node, result)
    return result


def _sort_node(node):
    line, context = node

    if context is not None:
        last = context.pop()
        context.sort()
        context.append(last)
        for next_node in context:
            _sort_node(next_node)


def print_node(node, result):
    line, context = node

    result.append(line)
    if context is not None:
        for next_node in context:
            print_node(next_node, result)


def get_approval() -> bool:
    answer = input()
    if answer != "yes" and answer != "y":
        return False
    else:
        return True



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import textwrap

from otterdog.config import OrganizationConfig
from otterdog.utils import jsonnet_evaluate_snippet, style

from . import Operation


class ShowDefaultOperation(Operation):
    """
    Shows the default configuration for organizations.
    """

    def __init__(self, markdown: bool):
        super().__init__()
        self._markdown = markdown

    @property
    def markdown(self) -> bool:
        return self._markdown

    def pre_execute(self) -> None:
        if not self.markdown:
            self.printer.println("Showing defaults configurations:")

    async def execute(self, org_config: OrganizationConfig) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config
        jsonnet_config.init_template()

        if not self.markdown:
            self.printer.println(f"\nOrganization {style(org_config.name, bright=True)}[id={github_id}]")
            self.printer.level_up()

        try:
            default_org = self.evaluate(jsonnet_config, f"{jsonnet_config.create_org}('<github-id>')")
            default_org_settings = {"settings": default_org["settings"]}
            self.printer.println()
            self._print_header("Organization Settings")
            self.print_dict(
                default_org_settings, f"orgs.{jsonnet_config.create_org}('<github-id>') =", "", "red", ":", ","
            )
            self._print_footer()

            default_org_webhook = self.evaluate(jsonnet_config, f"{jsonnet_config.create_org_webhook}('<url>')")
            self.printer.println()
            self._print_header("Organization Webhook")
            self.print_dict(
                default_org_webhook,
                f"orgs.{jsonnet_config.create_org_webhook}('<url>') =",
                "",
                "black",
                ":",
                ",",
            )
            self._print_footer()

            default_org_secret = self.evaluate(jsonnet_config, f"{jsonnet_config.create_org_secret}('<name>')")
            self.printer.println()
            self._print_header("Organization Secret")
            self.print_dict(
                default_org_secret,
                f"orgs.{jsonnet_config.create_org_secret}('<name>') =",
                "",
                "black",
                ":",
                ",",
            )
            self._print_footer()

            default_org_variable = self.evaluate(jsonnet_config, f"{jsonnet_config.create_org_variable}('<name>')")
            self.printer.println()
            self._print_header("Organization Variable")
            self.print_dict(
                default_org_variable,
                f"orgs.{jsonnet_config.create_org_variable}('<name>') =",
                "",
                "black",
                ":",
                ",",
            )
            self._print_footer()

            default_repo = self.evaluate(jsonnet_config, f"{jsonnet_config.create_repo}('<repo-name>')")
            self.printer.println()
            self._print_header("Repository")
            self.print_dict(default_repo, f"orgs.{jsonnet_config.create_repo}('<repo-name>') =", "", "black", ":", ",")
            self._print_footer()

            default_bpr = self.evaluate(jsonnet_config, f"{jsonnet_config.create_branch_protection_rule}('<pattern>')")
            self.printer.println()
            self._print_header("Branch Protection Rule")
            self.print_dict(
                default_bpr,
                f"orgs.{jsonnet_config.create_branch_protection_rule}('<pattern>') =",
                "",
                "black",
                ":",
                ",",
            )
            self._print_footer()

            default_ruleset = self.evaluate(jsonnet_config, f"{jsonnet_config.create_repo_ruleset}('<name>')")
            self.printer.println()
            self._print_header("Repository Ruleset")
            self.print_dict(
                default_ruleset,
                f"orgs.{jsonnet_config.create_repo_ruleset}('<name>') =",
                "",
                "black",
                ":",
                ",",
            )
            self._print_footer()

            default_repo_webhook = self.evaluate(jsonnet_config, f"{jsonnet_config.create_repo_webhook}('<url>')")
            self.printer.println()
            self._print_header("Repository Webhook")
            self.print_dict(
                default_repo_webhook,
                f"orgs.{jsonnet_config.create_repo_webhook}('<url>') =",
                "",
                "black",
                ":",
                ",",
            )
            self._print_footer()

            default_repo_secret = self.evaluate(jsonnet_config, f"{jsonnet_config.create_repo_secret}('<name>')")
            self.printer.println()
            self._print_header("Repository Secret")
            self.print_dict(
                default_repo_secret,
                f"orgs.{jsonnet_config.create_repo_secret}('<name>') =",
                "",
                "black",
                ":",
                ",",
            )
            self._print_footer()

            default_repo_variable = self.evaluate(jsonnet_config, f"{jsonnet_config.create_repo_variable}('<name>')")
            self.printer.println()
            self._print_header("Repository Variable")
            self.print_dict(
                default_repo_variable,
                f"orgs.{jsonnet_config.create_repo_variable}('<name>') =",
                "",
                "black",
                ":",
                ",",
            )
            self._print_footer()

            default_repo_env = self.evaluate(jsonnet_config, f"{jsonnet_config.create_environment}('<name>')")
            self.printer.println()
            self._print_header("Environment")
            self.print_dict(
                default_repo_env,
                f"orgs.{jsonnet_config.create_environment}('<name>') =",
                "",
                "black",
                ":",
                ",",
            )
            self._print_footer()

            return 0

        finally:
            if not self.markdown:
                self.printer.level_down()

    def _print_header(self, resource_name: str):
        if self.markdown:
            self.printer.println(
                textwrap.dedent(
                    f"""\
                === "{resource_name}"
                    ``` jsonnet\
                """
                )
            )
            self.printer.level_up()
            self.printer.level_up()

    def _print_footer(self):
        if self.markdown:
            self.printer.println("```")
            self.printer.level_down()
            self.printer.level_down()

    @staticmethod
    def evaluate(jsonnet_config, function: str):
        try:
            snippet = f"(import '{jsonnet_config.template_file}').{function}"
            return jsonnet_evaluate_snippet(snippet)
        except RuntimeError as ex:
            raise RuntimeError(f"failed to evaluate snippet: {ex}")



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, MutableMapping

from . import AuthImpl, AuthStrategy


@dataclass(frozen=True)
class TokenAuthStrategy(AuthStrategy):
    """
    An AuthStrategy using classic PATs.
    """

    token: str

    def get_auth(self) -> AuthImpl:
        return _TokenAuth(self.token)


@dataclass(frozen=True)
class _TokenAuth(AuthImpl):
    token: str

    def __call__(self, r):
        self.update_headers_with_authorization(r.headers)
        return r

    def update_headers_with_authorization(self, headers: MutableMapping[str, Any]) -> None:
        headers["Authorization"] = f"Bearer {self.token}"



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import os
from typing import Any, Optional

from otterdog.config import OrganizationConfig, OtterdogConfig
from otterdog.models import LivePatch, LivePatchType, ModelObject
from otterdog.utils import Change, IndentingPrinter, get_approval, style

from .diff_operation import DiffStatus
from .plan import PlanOperation


class ApplyOperation(PlanOperation):
    def __init__(
        self,
        force_processing: bool,
        no_web_ui: bool,
        update_webhooks: bool,
        update_secrets: bool,
        update_filter: str,
        delete_resources: bool,
        resolve_secrets: bool = True,
        include_resources_with_secrets: bool = True,
    ):
        super().__init__(no_web_ui, update_webhooks, update_secrets, update_filter)
        self._force_processing = force_processing
        self._delete_resources = delete_resources
        self._resolve_secrets = resolve_secrets
        self._include_resources_with_secrets = include_resources_with_secrets

    def init(self, config: OtterdogConfig, printer: IndentingPrinter) -> None:
        super().init(config, printer)

    def pre_execute(self) -> None:
        self.printer.println("Applying changes:")
        self.print_legend()

    def include_resources_with_secrets(self) -> bool:
        return self._include_resources_with_secrets

    def resolve_secrets(self) -> bool:
        return self._resolve_secrets

    def handle_add_object(
        self,
        org_id: str,
        model_object: ModelObject,
        parent_object: Optional[ModelObject] = None,
    ) -> None:
        super().handle_add_object(org_id, model_object, parent_object)
        self.execute_custom_hook_if_present(self.org_config, model_object, "pre-add-object-hook.py")

    def handle_delete_object(
        self,
        org_id: str,
        model_object: ModelObject,
        parent_object: Optional[ModelObject] = None,
    ) -> None:
        super().handle_delete_object(org_id, model_object, parent_object)

    def handle_modified_object(
        self,
        org_id: str,
        modified_object: dict[str, Change[Any]],
        forced_update: bool,
        current_object: ModelObject,
        expected_object: ModelObject,
        parent_object: Optional[ModelObject] = None,
    ) -> int:
        modified = super().handle_modified_object(
            org_id,
            modified_object,
            forced_update,
            current_object,
            expected_object,
            parent_object,
        )
        return modified

    async def handle_finish(self, org_id: str, diff_status: DiffStatus, patches: list[LivePatch]) -> None:
        self.printer.println()

        if diff_status.total_changes(self._delete_resources) == 0:
            self.printer.println("No changes required.")
            if not self._delete_resources and diff_status.deletions > 0:
                self.printer.println(
                    f"{diff_status.deletions} resource(s) would be deleted with flag '--delete-resources'."
                )
            return

        if not self._force_processing:
            if diff_status.deletions > 0 and not self._delete_resources:
                self.printer.println("No resource will be removed, use flag '--delete-resources' to delete them.\n")

            self.printer.println(
                "Do you want to perform these actions? (Only 'yes' or 'y' will be accepted to approve)\n"
            )

            self.printer.print(f"{style('Enter a value', bright=True)}: ")
            if not get_approval():
                self.printer.println("\nApply cancelled.")
                return

        # apply patches
        import click

        self.printer.println("\nApplying changes:\n")
        with click.progressbar(patches, file=self.printer.writer) as bar:
            for patch in bar:
                if patch.patch_type == LivePatchType.REMOVE and not self._delete_resources:
                    continue
                else:
                    try:
                        await patch.apply(org_id, self.gh_client)
                    except RuntimeError as ex:
                        self.printer.println()
                        self.printer.print_error(f"failed to apply patch: {patch!r}\n{ex}")

        delete_snippet = "deleted" if self._delete_resources else "live resources ignored"

        self.printer.println("Done.")

        self.printer.println(
            f"\n{style('Executed plan', bright=True)}: {diff_status.additions} added, "
            f"{diff_status.differences} changed, "
            f"{diff_status.deletions} {delete_snippet}."
        )

    def execute_custom_hook_if_present(
        self, org_config: OrganizationConfig, model_object: ModelObject, filename: str
    ) -> None:
        hook_script = os.path.join(self.template_dir, filename)
        if os.path.exists(hook_script):
            exec(open(hook_script).read())
